#
# Infopipe integration test
#
# Copyright (c) 2017 Red Hat, Inc.
# Author: Lukas Slebodnik <lslebodn@redhat.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import print_function

import os
import stat
import pwd
import signal
import subprocess
import errno
import time
import ldap
import ldap.modlist
import pytest
import dbus
import shutil

import config
import ds_openldap
import ldap_ent
from util import unindent, get_call_output

LDAP_BASE_DN = "dc=example,dc=com"
INTERACTIVE_TIMEOUT = 4


class DbusDaemon(object):
    def __init__(self):
        self.pid = 0

    def start(self):
        """Start the SSSD process"""
        assert self.pid == 0

        dbus_config_path = config.SYSCONFDIR + "/dbus-1/cwrap-dbus-system.conf"
        dbus_commands = [
            ["dbus-daemon", "--config-file", dbus_config_path,
             "--nosyslog", "--fork"],
            ["dbus-daemon", "--config-file", dbus_config_path, "--fork"],
        ]
        dbus_started = False
        for dbus_command in dbus_commands:
            try:
                if subprocess.call(dbus_command) == 0:
                    dbus_started = True
                    break
                else:
                    print("start failed for %s" % " ".join(dbus_command))
            except OSError as ex:
                if ex.errno == errno.ENOENT:
                    print("%s does not exist" % (dbus_command[0]))
                    pass

        if not dbus_started:
            raise Exception("dbus-daemon start failed")
        dbus_pid_path = config.RUNSTATEDIR + "/dbus/messagebus.pid"
        # wait 10 seconds for pidfile
        wait_time = 10
        for _ in range(wait_time * 10):
            if os.path.isfile(dbus_pid_path):
                break
            time.sleep(.1)

        assert os.path.isfile(dbus_pid_path)
        with open(dbus_pid_path, "r") as pid_file:
            self.pid = int(pid_file.read())

    def stop(self):
        """Stop the SSSD process and remove its state"""

        # stop process only if running
        if self.pid != 0:
            try:
                os.kill(self.pid, signal.SIGTERM)
                while True:
                    try:
                        os.kill(self.pid, signal.SIGCONT)
                    except OSError:
                        break
                    time.sleep(.1)
            except OSError:
                pass

        # clean pid so we can start service one more time
        self.pid = 0

        # dbus-daemon 1.2.24 does not clean pid file after itself
        try:
            os.unlink(config.RUNSTATEDIR + "/dbus/messagebus.pid")
        except OSError as ex:
            if ex.errno != errno.ENOENT:
                raise


@pytest.fixture(scope="module")
def dbus_system_bus(request):
    dbus_daemon = DbusDaemon()
    dbus_daemon.start()

    def cleanup_dbus_process():
        dbus_daemon.stop()
    request.addfinalizer(cleanup_dbus_process)

    return dbus.SystemBus()


@pytest.fixture(scope="module")
def ds_inst(request):
    """LDAP server instance fixture"""
    ds_inst = ds_openldap.DSOpenLDAP(
        config.PREFIX, 10389, LDAP_BASE_DN,
        "cn=admin", "Secret123"
    )

    try:
        ds_inst.setup()
    except Exception:
        ds_inst.teardown()
        raise
    request.addfinalizer(ds_inst.teardown)

    return ds_inst


@pytest.fixture(scope="module")
def ldap_conn(request, ds_inst):
    """LDAP server connection fixture"""
    ldap_conn = ds_inst.bind()
    ldap_conn.ds_inst = ds_inst
    request.addfinalizer(ldap_conn.unbind_s)
    return ldap_conn


def create_ldap_entries(ldap_conn, ent_list=None):
    """Add LDAP entries from ent_list"""
    if ent_list is not None:
        for entry in ent_list:
            ldap_conn.add_s(entry[0], entry[1])


def cleanup_ldap_entries(ldap_conn, ent_list=None):
    """Remove LDAP entries added by create_ldap_entries"""
    if ent_list is None:
        for ou in ("Users", "Groups", "Netgroups", "Services", "Policies"):
            for entry in ldap_conn.search_s(f"ou={ou},"
                                            f"{ldap_conn.ds_inst.base_dn}",
                                            ldap.SCOPE_ONELEVEL,
                                            attrlist=[]):
                ldap_conn.delete_s(entry[0])
    else:
        for entry in ent_list:
            ldap_conn.delete_s(entry[0])


def create_ldap_cleanup(request, ldap_conn, ent_list=None):
    """Add teardown for removing all user/group LDAP entries"""
    request.addfinalizer(lambda: cleanup_ldap_entries(ldap_conn, ent_list))


def create_ldap_fixture(request, ldap_conn, ent_list=None):
    """Add LDAP entries and add teardown for removing them"""
    create_ldap_entries(ldap_conn, ent_list)
    create_ldap_cleanup(request, ldap_conn, ent_list)


SCHEMA_RFC2307 = "rfc2307"
SCHEMA_RFC2307_BIS = "rfc2307bis"


def format_basic_conf(ldap_conn, schema, config):
    """Format a basic SSSD configuration"""
    schema_conf = "ldap_schema         = " + schema + "\n"
    if schema == SCHEMA_RFC2307_BIS:
        schema_conf += "ldap_group_object_class = groupOfNames\n"

    return unindent("""\
        [sssd]
        debug_level         = 0xffff
        domains             = LDAP, app
        services            = nss, ifp
        enable_files_domain = false

        [nss]
        memcache_timeout    = 0

        [ifp]
        debug_level         = 0xffff
        user_attributes = +extraName
        ca_db               = {config.PAM_CERT_DB_PATH}

        [domain/LDAP]
        {schema_conf}
        id_provider         = ldap
        ldap_uri            = {ldap_conn.ds_inst.ldap_url}
        ldap_search_base    = {ldap_conn.ds_inst.base_dn}
        ldap_id_use_start_tls = false
        ldap_user_extra_attrs = extraName:uid
        ldap_user_certificate = userCert

        [application/app]
        inherit_from = LDAP
    """).format(**locals())


def format_certificate_conf(ldap_conn, schema, config):
    """Format an SSSD configuration with all caches refreshing in 4 seconds"""
    return \
        format_basic_conf(ldap_conn, schema, config) + \
        unindent("""
            [certmap/LDAP/user1]
            matchrule = <SUBJECT>.*CN = SSSD test cert 0001.*
        """).format(**locals())


def create_conf_file(contents):
    """Create sssd.conf with specified contents"""
    with open(config.CONF_PATH, "w") as conf:
        conf.write(contents)
    os.chmod(config.CONF_PATH, stat.S_IRUSR | stat.S_IWUSR)


def cleanup_conf_file():
    """Remove sssd.conf, if it exists"""
    if os.path.lexists(config.CONF_PATH):
        os.unlink(config.CONF_PATH)


def create_conf_cleanup(request):
    """Add teardown for removing sssd.conf"""
    request.addfinalizer(cleanup_conf_file)


def create_conf_fixture(request, contents):
    """
    Create sssd.conf with specified contents and add teardown for removing it
    """
    create_conf_file(contents)
    create_conf_cleanup(request)


def create_sssd_process():
    """Start the SSSD process"""
    if subprocess.call(["sssd", "-D", "--logger=files"]) != 0:
        raise Exception("sssd start failed")


def cleanup_sssd_process():
    """Stop the SSSD process and remove its state"""
    try:
        with open(config.PIDFILE_PATH, "r") as pid_file:
            pid = int(pid_file.read())
        os.kill(pid, signal.SIGTERM)
        while True:
            try:
                os.kill(pid, signal.SIGCONT)
            except OSError:
                break
            time.sleep(1)
    except OSError:
        pass
    for path in os.listdir(config.DB_PATH):
        os.unlink(config.DB_PATH + "/" + path)
    for path in os.listdir(config.MCACHE_PATH):
        os.unlink(config.MCACHE_PATH + "/" + path)


def create_sssd_cleanup(request):
    """Add teardown for stopping SSSD and removing its state"""
    request.addfinalizer(cleanup_sssd_process)


def create_sssd_fixture(request):
    """Start SSSD and add teardown for stopping it and removing its state"""
    create_sssd_process()
    create_sssd_cleanup(request)


def backup_ca_db():
    """Create backup file for ca db"""
    src = os.path.dirname(config.PAM_CERT_DB_PATH) + "/SSSD_test_CA.pem"
    dst = os.path.dirname(config.PAM_CERT_DB_PATH) + "/SSSD_test_CA.pem.bp"
    shutil.copyfile(src, dst)


def restore_ca_db():
    """Restore backup file for ca db"""
    src = os.path.dirname(config.PAM_CERT_DB_PATH) + "/SSSD_test_CA.pem.bp"
    dst = os.path.dirname(config.PAM_CERT_DB_PATH) + "/SSSD_test_CA.pem"
    shutil.copyfile(src, dst)
    os.remove(src)


def create_restore_ca_db(request):
    """Add teardown for restoring ca_db"""
    request.addfinalizer(restore_ca_db)


def create_ca_db_fixture(request):
    """
    Create backup for ca_db and add teardown for restoring it
    """
    backup_ca_db()
    create_restore_ca_db(request)


@pytest.fixture
def sanity_rfc2307(request, ldap_conn):
    ent_list = ldap_ent.List(ldap_conn.ds_inst.base_dn)
    ent_list.add_user("user1", 1001, 2001)
    ent_list.add_user("user2", 1002, 2002)
    ent_list.add_user("user3", 1003, 2003)

    ent_list.add_group("group1", 2001)
    ent_list.add_group("group2", 2002)
    ent_list.add_group("group3", 2003)

    ent_list.add_group("empty_group", 2010)

    ent_list.add_group("single_user_group", 2011, ["user1"])
    ent_list.add_group("two_user_group", 2012, ["user1", "user2"])

    create_ldap_fixture(request, ldap_conn, ent_list)

    config.PAM_CERT_DB_PATH = os.environ['PAM_CERT_DB_PATH']
    conf = format_basic_conf(ldap_conn, SCHEMA_RFC2307, config)
    create_conf_fixture(request, conf)
    create_sssd_fixture(request)
    create_ca_db_fixture(request)
    return None


@pytest.fixture
def simple_rfc2307(request, ldap_conn):
    ent_list = ldap_ent.List(ldap_conn.ds_inst.base_dn)
    ent_list.add_user('usr\\\\001', 181818, 181818)
    ent_list.add_group("group1", 181818)
    create_ldap_fixture(request, ldap_conn, ent_list)
    config.PAM_CERT_DB_PATH = os.environ['PAM_CERT_DB_PATH']
    conf = format_basic_conf(ldap_conn, SCHEMA_RFC2307, config)
    create_conf_fixture(request, conf)
    create_sssd_fixture(request)
    create_ca_db_fixture(request)
    return None


@pytest.fixture
def auto_private_groups_rfc2307(request, ldap_conn):
    ent_list = ldap_ent.List(ldap_conn.ds_inst.base_dn)
    ent_list.add_user("user1", 1001, 2001)

    ent_list.add_group("group1", 2001)
    ent_list.add_group("single_user_group", 2011, ["user1"])
    ent_list.add_group("two_user_group", 2012, ["user1"])

    create_ldap_fixture(request, ldap_conn, ent_list)

    config.PAM_CERT_DB_PATH = os.environ['PAM_CERT_DB_PATH']
    conf = \
        format_basic_conf(ldap_conn, SCHEMA_RFC2307, config) + \
        unindent("""
            [domain/LDAP]
            auto_private_groups = True
        """).format(**locals())
    create_conf_fixture(request, conf)
    create_sssd_fixture(request)
    create_ca_db_fixture(request)
    return None


@pytest.fixture
def add_user_with_cert(request, ldap_conn):
    config.PAM_CERT_DB_PATH = os.environ['PAM_CERT_DB_PATH']

    ent_list = ldap_ent.List(ldap_conn.ds_inst.base_dn)
    ent_list.add_user("user1", 1001, 2001)

    create_ldap_fixture(request, ldap_conn, ent_list)

    der_path = os.path.dirname(config.PAM_CERT_DB_PATH)
    der_path += "/SSSD_test_cert_x509_0001.der"
    with open(der_path, 'rb') as f:
        val = f.read()
    dn = "uid=user1,ou=Users," + LDAP_BASE_DN
    '''
    Using 'userCert' instead of 'userCertificate' to hold the user certificate
    because the default OpenLDAP has syntax and matching rules which are not
    used in other LDAP servers.
    '''
    ldap_conn.modify_s(dn, [(ldap.MOD_ADD, 'userCert', val)])

    conf = format_certificate_conf(ldap_conn, SCHEMA_RFC2307_BIS, config)
    create_conf_fixture(request, conf)
    create_sssd_fixture(request)
    create_ca_db_fixture(request)

    return None


def test_ping_raw(dbus_system_bus, ldap_conn, simple_rfc2307):
    # test with disabled introspection
    sssd_obj = dbus_system_bus.get_object('org.freedesktop.sssd.infopipe',
                                          '/org/freedesktop/sssd/infopipe',
                                          introspect=False)
    sssd_interface = dbus.Interface(sssd_obj, 'org.freedesktop.sssd.infopipe')

    # test missing parameter
    with pytest.raises(dbus.exceptions.DBusException) as exc_info:
        sssd_interface.Ping()
    assert exc_info.errisinstance(dbus.exceptions.DBusException)

    ex = exc_info.value
    assert ex.get_dbus_name() == 'sbus.Error.Errno'
    assert 'Unexpected argument type provided' in ex.get_dbus_message()

    # test wrong parameter type
    with pytest.raises(dbus.exceptions.DBusException) as exc_info:
        sssd_interface.Ping(1)
    assert exc_info.errisinstance(dbus.exceptions.DBusException)

    ex = exc_info.value
    assert ex.get_dbus_name() == 'sbus.Error.Errno'
    assert 'Unexpected argument type provided' in ex.get_dbus_message()

    # test wrong parameter value
    with pytest.raises(dbus.exceptions.DBusException) as exc_info:
        sssd_interface.Ping('test')
    assert exc_info.errisinstance(dbus.exceptions.DBusException)

    ex = exc_info.value
    assert ex.get_dbus_name() == 'org.freedesktop.DBus.Error.InvalidArgs'
    assert ex.get_dbus_message() == 'Invalid argument'

    # positive test
    ret = sssd_interface.Ping('ping')
    assert ret == "PONG"

    # test case insensitive input
    ret = sssd_interface.Ping('PinG')
    assert ret == "PONG"

    ret = sssd_interface.Ping('PING')
    assert ret == "PONG"


def test_ping_introspection(dbus_system_bus, ldap_conn, simple_rfc2307):
    sssd_obj = dbus_system_bus.get_object('org.freedesktop.sssd.infopipe',
                                          '/org/freedesktop/sssd/infopipe')
    sssd_interface = dbus.Interface(sssd_obj, 'org.freedesktop.sssd.infopipe')

    # test missing parameter
    with pytest.raises(TypeError) as exc_info:
        sssd_interface.Ping()
    assert exc_info.errisinstance(TypeError)

    ex = exc_info.value
    assert str(ex) == 'More items found in D-Bus signature than in Python ' \
                      'arguments'

    # test wrong parameter type
    with pytest.raises(TypeError) as exc_info:
        sssd_interface.Ping(1)
    assert exc_info.errisinstance(TypeError)

    ex = exc_info.value
    assert str(ex) == 'Expected a string or unicode object'

    # test wrong parameter value
    with pytest.raises(dbus.exceptions.DBusException) as exc_info:
        sssd_interface.Ping('test')
    assert exc_info.errisinstance(dbus.exceptions.DBusException)

    ex = exc_info.value
    assert ex.get_dbus_name() == 'org.freedesktop.DBus.Error.InvalidArgs'
    assert ex.get_dbus_message() == 'Invalid argument'

    # positive test
    ret = sssd_interface.Ping('ping')
    assert ret == "PONG"

    # test case insensitive input
    ret = sssd_interface.Ping('PinG')
    assert ret == "PONG"

    ret = sssd_interface.Ping('PING')
    assert ret == "PONG"


def test_special_characters(dbus_system_bus, ldap_conn, simple_rfc2307):
    sssd_obj = dbus_system_bus.get_object('org.freedesktop.sssd.infopipe',
                                          '/org/freedesktop/sssd/infopipe')
    sssd_interface = dbus.Interface(sssd_obj, 'org.freedesktop.sssd.infopipe')

    attributes = ['name', 'uidNumber', 'gidNumber', 'gecos', 'homeDirectory',
                  'loginShell']
    expected = dict(name='usr\\001', uidNumber='181818', gidNumber='181818',
                    gecos='181818', homeDirectory='/home/usr\\\\001',
                    loginShell='/bin/bash')

    user_attrs = sssd_interface.GetUserAttr('usr\\001', attributes)
    assert user_attrs.signature == 'sv'
    assert user_attrs.variant_level == 0

    assert len(attributes) == len(user_attrs)
    assert sorted(attributes) == sorted(user_attrs.keys())

    # check values of attributes
    for attr in user_attrs:
        assert user_attrs[attr].signature == 's'
        assert user_attrs[attr].variant_level == 1
        assert user_attrs[attr][0] == expected[attr]


def test_get_user_attr(dbus_system_bus, ldap_conn, sanity_rfc2307):
    sssd_obj = dbus_system_bus.get_object('org.freedesktop.sssd.infopipe',
                                          '/org/freedesktop/sssd/infopipe')
    sssd_interface = dbus.Interface(sssd_obj, 'org.freedesktop.sssd.infopipe')

    # negative test
    with pytest.raises(dbus.exceptions.DBusException) as exc_info:
        sssd_interface.GetUserAttr('non_existent_user', ['name'])
    assert exc_info.errisinstance(dbus.exceptions.DBusException)

    ex = exc_info.value
    assert ex.get_dbus_name() == 'sbus.Error.NotFound'
    assert ex.get_dbus_message() == 'No such file or directory'

    # test 0 attributes
    user_attrs = sssd_interface.GetUserAttr('user1', [])

    assert user_attrs.signature == 'sv'
    assert user_attrs.variant_level == 0

    # expect empty sequence; len(user_attrs) == 0
    assert not user_attrs

    # positive test
    attributes = ['name', 'uidNumber', 'gidNumber', 'gecos', 'homeDirectory',
                  'loginShell']
    expected = dict(name='user1', uidNumber='1001', gidNumber='2001',
                    gecos='1001', homeDirectory='/home/user1',
                    loginShell='/bin/bash')
    user_attrs = sssd_interface.GetUserAttr('user1', attributes)

    assert user_attrs.signature == 'sv'
    assert user_attrs.variant_level == 0

    assert len(attributes) == len(user_attrs)
    assert sorted(attributes) == sorted(user_attrs.keys())

    # check values of attributes
    for attr in user_attrs:
        assert user_attrs[attr].signature == 's'
        assert user_attrs[attr].variant_level == 1
        assert user_attrs[attr][0] == expected[attr]


def test_get_user_groups(dbus_system_bus, ldap_conn, sanity_rfc2307):
    sssd_obj = dbus_system_bus.get_object('org.freedesktop.sssd.infopipe',
                                          '/org/freedesktop/sssd/infopipe')
    sssd_interface = dbus.Interface(sssd_obj, 'org.freedesktop.sssd.infopipe')

    # negative test
    with pytest.raises(dbus.exceptions.DBusException) as exc_info:
        sssd_interface.GetUserGroups('non_existent_user')
    assert exc_info.errisinstance(dbus.exceptions.DBusException)

    ex = exc_info.value
    assert ex.get_dbus_name() == 'sbus.Error.NotFound'
    assert ex.get_dbus_message() == 'No such file or directory'

    # the same test via nss responder
    with pytest.raises(KeyError):
        pwd.getpwnam("non_existent_user")

    # 0 groups
    res = sssd_interface.GetUserGroups('user3')
    assert res.signature == 's'
    assert res.variant_level == 0

    # expect empty sequence; len(res) == 0
    assert not res

    # single group
    res = sssd_interface.GetUserGroups('user2')
    assert res.signature == 's'
    assert res.variant_level == 0

    assert len(res) == 1
    assert res[0] == 'two_user_group'

    # more groups
    res = sssd_interface.GetUserGroups('user1')
    assert res.signature == 's'
    assert res.variant_level == 0

    assert len(res) == 2
    assert sorted(res) == ['single_user_group', 'two_user_group']


'''
Given auto_private_groups is enabled
When GetUserGroups is called
Then the origPrimaryGroupGidNumber is returned as part of the group memberships
'''


def test_get_user_groups_given_auto_private_groups_enabled(
        dbus_system_bus,
        ldap_conn, auto_private_groups_rfc2307):
    sssd_obj = dbus_system_bus.get_object('org.freedesktop.sssd.infopipe',
                                          '/org/freedesktop/sssd/infopipe')
    sssd_interface = dbus.Interface(sssd_obj, 'org.freedesktop.sssd.infopipe')

    res = sssd_interface.GetUserGroups('user1')

    assert sorted(res) == ['group1', 'single_user_group', 'two_user_group']


def get_user_property(dbus_system_bus, username, prop_name):
    users_obj = dbus_system_bus.get_object('org.freedesktop.sssd.infopipe',
                                           '/org/freedesktop/sssd/infopipe/Users')

    users_iface = dbus.Interface(users_obj,
                                 "org.freedesktop.sssd.infopipe.Users")

    user_path = users_iface.FindByName(username)
    user_object = dbus_system_bus.get_object('org.freedesktop.sssd.infopipe',
                                             user_path)

    prop_iface = dbus.Interface(user_object, 'org.freedesktop.DBus.Properties')
    return prop_iface.Get('org.freedesktop.sssd.infopipe.Users.User',
                          prop_name)


def get_user_by_attr(dbus_system_bus, attribute, filter):
    users_obj = dbus_system_bus.get_object('org.freedesktop.sssd.infopipe',
                                           '/org/freedesktop/sssd/infopipe/Users')

    users_iface = dbus.Interface(users_obj,
                                 "org.freedesktop.sssd.infopipe.Users")

    return users_iface.ListByAttr(attribute, filter, 0)


def get_user_by_name(dbus_system_bus, filter):
    users_obj = dbus_system_bus.get_object('org.freedesktop.sssd.infopipe',
                                           '/org/freedesktop/sssd/infopipe/Users')

    users_iface = dbus.Interface(users_obj,
                                 "org.freedesktop.sssd.infopipe.Users")

    return users_iface.ListByName(filter, 0)


def test_get_extra_attributes_empty(dbus_system_bus,
                                    ldap_conn,
                                    sanity_rfc2307):
    """
    Make sure the extraAttributes property can be retrieved
    """
    extra_attrs = get_user_property(dbus_system_bus,
                                    'user1',
                                    'extraAttributes')
    assert extra_attrs['extraName'][0] == 'user1'


def test_sssctl_domain_list_app_domain(dbus_system_bus,
                                       ldap_conn,
                                       sanity_rfc2307):
    output = get_call_output(["sssctl", "domain-list"], subprocess.STDOUT)

    assert "Error" not in output
    assert output.find("LDAP") != -1
    assert output.find("app") != -1


def test_update_member_list_and_get_all(dbus_system_bus,
                                        ldap_conn,
                                        sanity_rfc2307):
    '''
    Test that UpdateMemberList() and GetAll() return the correct users that are
    members of a group
    '''
    sssd_obj = dbus_system_bus.get_object(
        'org.freedesktop.sssd.infopipe',
        '/org/freedesktop/sssd/infopipe/Groups')
    groups_iface = dbus.Interface(sssd_obj,
                                  'org.freedesktop.sssd.infopipe.Groups')
    group_id = 2011
    expected_user_result = "/org/freedesktop/sssd/infopipe/Users/LDAP/1001"

    group_path = groups_iface.FindByName('single_user_group')

    group_object = dbus_system_bus.get_object('org.freedesktop.sssd.infopipe',
                                              group_path)
    group_iface = dbus.Interface(group_object,
                                 'org.freedesktop.sssd.infopipe.Groups.Group')

    # update local cache for group
    try:
        group_iface.UpdateMemberList(group_id)
    except dbus.exceptions.DBusException as ex:
        assert False, "Unexpected DBusException raised: " + ex

    # check members of group
    prop_iface = dbus.Interface(group_object,
                                'org.freedesktop.DBus.Properties')
    res = prop_iface.GetAll('org.freedesktop.sssd.infopipe.Groups.Group')
    assert str(res.get("users")[0]) == expected_user_result

    # delete group (there's no other way of removing a user from a group) and
    # wait change to propagate
    ldap_conn.delete("cn=single_user_group,ou=Groups,dc=example,dc=com")
    time.sleep(INTERACTIVE_TIMEOUT)

    # add group back but this time without any member
    ldap_group = ldap_ent.group("dc=example,dc=com", "single_user_group", 2011)
    ldap_conn.add_s(ldap_group[0], ldap_group[1])

    # invalidate cache
    subprocess.call(["sss_cache", "-E"])

    # check that group has no members
    group_iface.UpdateMemberList(group_id)
    prop_interface = dbus.Interface(group_object,
                                    'org.freedesktop.DBus.Properties')
    res = prop_interface.GetAll('org.freedesktop.sssd.infopipe.Groups.Group')
    assert not res.get("users")


def test_find_by_valid_certificate(dbus_system_bus,
                                   ldap_conn,
                                   add_user_with_cert):
    """test_find_by_valid_certificate

    :id: 3f212e6e-00ce-44ac-95d4-59925cb5a14a
    :title: SSSD-TC: Infopipe: Find by valid certificate
    :casecomponent: sssd
    :subsystemteam: sst_idm_sssd
    """
    users_obj = dbus_system_bus.get_object(
        'org.freedesktop.sssd.infopipe',
        '/org/freedesktop/sssd/infopipe/Users')
    users_iface = dbus.Interface(users_obj,
                                 'org.freedesktop.sssd.infopipe.Users')
    cert_path = os.path.dirname(config.PAM_CERT_DB_PATH)

    # Valid certificate with user
    cert_file = cert_path + "/SSSD_test_cert_x509_0001.pem"
    with open(cert_file, "r") as f:
        cert = f.read()
    res = users_iface.FindByValidCertificate(cert)
    assert res == "/org/freedesktop/sssd/infopipe/Users/app/user1_40app"

    # Valid certificate without user
    cert_file = cert_path + "/SSSD_test_cert_x509_0002.pem"
    with open(cert_file, "r") as f:
        cert = f.read()
    try:
        res = users_iface.FindByValidCertificate(cert)
        assert False, "Previous call should raise an exception"
    except dbus.exceptions.DBusException as ex:
        assert str(ex) == "sbus.Error.NotFound: No such file or directory"

    # Valid certificate from another CA
    cert_file = os.environ['ABS_SRCDIR'] + \
        "/../test_ECC_CA/SSSD_test_ECC_cert_key_0001.pem"
    with open(cert_file, "r") as f:
        cert = f.read()
    try:
        res = users_iface.FindByValidCertificate(cert)
        assert False, "Previous call should raise an exception"
    except dbus.exceptions.DBusException as ex:
        assert str(ex) == \
            "org.freedesktop.DBus.Error.IOError: Input/output error"

    # Invalid certificate
    cert = "Invalid cert"
    try:
        res = users_iface.FindByValidCertificate(cert)
        assert False, "Previous call should raise an exception"
    except dbus.exceptions.DBusException as ex:
        error = "org.freedesktop.DBus.Error.IOError: Input/output error"
        assert str(ex) == error

    # Remove certificate db
    cert_db = cert_path + "/SSSD_test_CA.pem"
    os.remove(cert_db)
    cert_file = cert_path + "/SSSD_test_cert_x509_0002.pem"
    with open(cert_file, "r") as f:
        cert = f.read()
    try:
        res = users_iface.FindByValidCertificate(cert)
        assert False, "Previous call should raise an exception"
    except dbus.exceptions.DBusException as ex:
        assert str(ex) == \
            "sbus.Error.NoCA: Certificate authority file not found"



"""This module contains methods to create Directory Server Instance."""
from __future__ import print_function
import os
try:
    import ConfigParser
except ImportError:
    import configparser as ConfigParser
import tempfile
import subprocess
import socket
import time
import ldap
from sssd.testlib.common.exceptions import DirSrvException
from sssd.testlib.common.exceptions import LdapException
from sssd.testlib.common.utils import LdapOperations

DS_USER = 'dirsrv'
DS_GROUP = 'dirsrv'
DS_ADMIN = 'admin'
DS_ROOTDN = 'cn=Directory Manager'


class DirSrv(object):
    """Base class to setup DS Instances

    For setting up Directory Server, enabling TLS, and
    removing of Directory Server instance.
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, **kwargs):
        """Initialize name, suffix, host, ports, Directory Manager password."""
        self.instance_name = kwargs.get('name')
        self.dsinstance_host = kwargs.get('host')
        self.dsinstance_suffix = kwargs.get('suffix')
        self.multihost = kwargs.get('multihost')
        self.dsrootdn_pwd = kwargs.get('root_dn_password')
        self.dsldap_port = kwargs.get('ldap_port', None)
        self.dstls_port = kwargs.get('ldap_tls_port', None)
        self.dsrootdn = DS_ROOTDN
        self.ds_inst_name = 'slapd-%s' % self.instance_name
        self.dsrootdir = '/etc/dirsrv'
        self.dsinst_path = os.path.join(self.dsrootdir, self.ds_inst_name)

    def __str__(self):
        return "%s.%s('%r')" % (self.__module__, self.__class__.__name__,
                                self.__dict__)

    def __repr__(self):
        return '%s(%s, %r)' % (self.__module__, self.__class__.__name__,
                               self.__dict__)

    def create_config(self):
        """create inf file for dscreate to setup DS instances.

        Args:
            param1 (None):

        Returns:
            str: Returns path of the config file

        Exceptions:
            None
        """
        config = ConfigParser.RawConfigParser()
        config.optionxform = str
        config.add_section('general')
        config.set('general', 'full_machine_name', self.dsinstance_host)
        config.set('general', 'user', DS_USER)
        config.set('general', 'group', DS_GROUP)
        config.add_section('slapd')
        config.set('slapd', 'instance_name', self.instance_name)
        config.set('slapd', 'port', self.dsldap_port)
        config.set('slapd', 'suffix', self.dsinstance_suffix)
        config.set('slapd', 'root_dn', self.dsrootdn)
        config.set('slapd', 'root_password', self.dsrootdn_pwd)
        config.set('slapd', 'sample_entries', "yes")
        config.set('slapd', 'self_sign_cert', False)
        config.add_section('backend-userRoot')
        config.set('backend-userRoot', 'create_suffix_entry', 'True')
        config.set('backend-userRoot', 'sample_entries', '001003006')
        config.set('backend-userRoot', 'suffix', self.dsinstance_suffix)
        (ds_config, ds_config_file_path) = tempfile.mkstemp(suffix='cfg')
        os.close(ds_config)
        with open(ds_config_file_path, "w") as outfile:
            config.write(outfile)
        return ds_config_file_path

    def setup_ds(self, ds_cfg_file):
        """create DS instance by running dscreate

        Args:
             ds_config_file (str): ds_config_file: Configuration File path

        Returns:
             bool: True if dscreate ran successfully else False

        Exceptions:
             subprocess.CalledProcessError:
        """
        self.multihost.transport.put_file(ds_cfg_file, '/tmp/test.cfg')
        setup_cmd = 'dscreate -v from-file %s' % '/tmp/test.cfg'
        try:
            self.multihost.run_command(setup_cmd)
        except subprocess.CalledProcessError:
            self.multihost.log.info("Failed to setup Directory Server")
            raise
        else:
            os.remove(ds_cfg_file)
            return True

    def remove_ds(self, inst_name=None):
        """Remove Directory Server instance

        Args:
            inst_name (str): DS Instance name

        Returns:
            None

        Exceptions:
            subprocess.CalledProcessError
        """
        if inst_name is None:
            inst_name = self.ds_inst_name
        remove_cmd = 'dsctl %s remove --do-it' % (inst_name)
        try:
            self.multihost.run_command(remove_cmd)
        except subprocess.CalledProcessError:
            self.multihost.log.info("Failed to remove %s instance" % inst_name)
            raise

    def _copy_pkcs12(self, ssl_dir):
        """ Copy the pkcs12 files from ssl_dir to
        DS instance directory """

        server_p12 = '%s-server.p12' % self.multihost.sys_hostname
        nss_db_files = ['ca.p12', 'pin.txt', 'pwfile', server_p12]
        for db_file in nss_db_files:
            source = os.path.join(ssl_dir, db_file)
            destination = os.path.join(self.dsinst_path, db_file)
            self.multihost.transport.put_file(source, destination)
        for db_file in nss_db_files:
            ls_cmd = 'ls %s/%s' % (self.dsinst_path, db_file)
            cmd = self.multihost.run_command(ls_cmd)
            if cmd.returncode != 0:
                return False
        return True

    def _import_certs(self, pkcs12_path, pwfile):
        """ Import the certs from pkcs12 """
        pk12_cmd = 'pk12util -i %s -d %s -k %s'\
                   ' -w %s' % (pkcs12_path, self.dsinst_path, pwfile, pwfile)
        cmd = self.multihost.run_command(pk12_cmd)
        if cmd.returncode == 0:
            return True

    def _set_dsperms(self, file_path):
        """ Set DSUSER permissions on files """
        change_ownership = ['chown', DS_USER, file_path]
        change_group = ['chgrp', DS_GROUP, file_path]
        chmod_file = ['chmod', '600', file_path]
        try:
            self.multihost.run_command(change_ownership)
        except subprocess.CalledProcessError:
            raise DirSrvException(
                'Failed to change ownership of pin.txt')
        try:
            self.multihost.run_command(change_group)
        except subprocess.CalledProcessError:
            raise DirSrvException(
                'Failed to change group ownership of pin.txt')
        try:
            self.multihost.run_command(chmod_file)
        except subprocess.CalledProcessError:
            raise DirSrvException('Failed to change permissions of pin.txt')

    def setup_certs(self, ssl_dir, client_host=None, canick=None):
        """copy CA and Server certs to all DS instances.

        Args:
            ssl_dir (str): NSS Directory containing CA and Server-Certs

        Returns:
            bool: True if files are copied

        Exceptions:
            DirSrvException
        """
        # We stop directory server before we copy files. This is required
        # because it's seen that at times, if ns-slapd process is reading
        # the db files, copying of files is successful but not all data
        # is written causing the files to go corrupt.
        stop_ds = ['systemctl', 'stop', 'dirsrv@%s' % (self.instance_name)]
        try:
            self.multihost.run_command(stop_ds)
        except subprocess.CalledProcessError:
            raise DirSrvException("Failed to stop Directory Server instance")
        else:
            self.multihost.log.info('DS instance stopped successfully')
            self._copy_pkcs12(ssl_dir)
        cacert_file_path = '%s/cacert.pem' % ('/etc/openldap/cacerts')
        target_pin_file = os.path.join(self.dsinst_path, 'pin.txt')
        pwfile = os.path.join(self.dsinst_path, 'pwfile')
        ca_p12 = os.path.join(self.dsinst_path, 'ca.p12')
        server_p12_name = '%s-%s' % (self.multihost.sys_hostname, 'server.p12')
        server_p12 = os.path.join(self.dsinst_path, server_p12_name)
        # recreate the database
        certutil_cmd = 'certutil -T -d %s -f %s' % (self.dsinst_path, pwfile)
        self.multihost.run_command(certutil_cmd)
        create_cert_dir = 'mkdir -p /etc/openldap/cacerts'
        # create directory to copy ca cert
        self.multihost.run_command(create_cert_dir)
        if not canick:
            canick = "ExampleCA"
        pkcs12_file = [ca_p12, server_p12]
        for pkcs_file in pkcs12_file:
            if not self._import_certs(pkcs_file, pwfile):
                raise DirSrvException("importing certificates failed")
        set_trust_cmd = 'certutil -M -d %s -n %s ' \
                        '-t "CT,C,T" -f %s' % (self.dsinst_path,
                                               canick, pwfile)
        self.multihost.run_command(create_cert_dir)
        self.multihost.run_command(set_trust_cmd)
        self.multihost.transport.put_file(os.path.join(
            ssl_dir, 'cacert.pem'), cacert_file_path)
        if client_host:
            client_host.run_command(create_cert_dir)
            client_host.transport.put_file(os.path.join(
                ssl_dir, 'cacert.pem'), cacert_file_path)
        try:
            self._set_dsperms(target_pin_file)
        except DirSrvException:
            raise
        start_ds = ['systemctl', 'start', 'dirsrv@%s' % (self.instance_name)]
        try:
            self.multihost.run_command(start_ds)
        except subprocess.CalledProcessError:
            raise DirSrvException('Failed to start DS Instance')
        else:
            self.multihost.log.info('DS instance started successfully')

    def enable_anonymous_search(self, binduri):
        """Enable anonymous search access to basedn
        Args:
            binduri (str): LDAP uri to bind with
        Returns:
            boold: True if ACI is added
        Exceptions:
            LdapException
        """
        ldap_obj = LdapOperations(uri=binduri, binddn=self.dsrootdn,
                                  bindpw=self.dsrootdn_pwd)
        # Enable Anonymous access aci
        allow_anonymous = "(targetattr!=\"userPassword || aci\")" \
                          "(version 3.0; acl \"Enable anonymous access\";" \
                          "allow (read, search, compare)" \
                          "userdn=\"ldap:///anyone\";)"
        add_aci = [(ldap.MOD_ADD, 'aci', [allow_anonymous.encode('utf-8')])]
        (ret, return_value) = ldap_obj.modify_ldap(self.dsinstance_suffix,
                                                   add_aci)
        if not return_value:
            raise LdapException("Failed to enable anonymous access aci")
        else:
            print("Enabled Anonymous access aci to %s" %
                  self.dsinstance_suffix)

    def enable_ssl(self, binduri, tls_port):
        """sets TLS Port and enabled TLS on Directory Server.

        Args:
            binduri (str): LDAP uri to bind with
            tls_port (str): TLS port to be setup

        Returns:
            bool: True if successfully setup TLS port

        Exceptions:
            LdapException
        """
        ldap_obj = LdapOperations(uri=binduri, binddn=self.dsrootdn,
                                  bindpw=self.dsrootdn_pwd)
        # Enable TLS
        mod_dn1 = 'cn=encryption,cn=config'
        add_tls = [(ldap.MOD_ADD, 'nsTLS1', [b'on'])]
        (ret, return_value) = ldap_obj.modify_ldap(mod_dn1, add_tls)
        if not return_value:
            raise LdapException('Failed to enable TLS, Error:%s' % (ret))
        else:
            print('Enabled nsTLS1=on')
        mod_dn2 = 'cn=RSA,cn=encryption,cn=config'
        mod_security = [(ldap.MOD_REPLACE, 'nsSSLPersonalitySSL',
                         [b'Server-Cert-%s' %
                          ((self.dsinstance_host.encode()))])]
        (ret, return_value) = ldap_obj.modify_ldap(mod_dn2, mod_security)
        if not return_value:
            raise LdapException('Failed to set Server-Cert nick:%s' % (ret))
        else:
            print('Enabled Server-Cert nick')

        # Enable security
        mod_dn3 = 'cn=config'
        enable_security = [(ldap.MOD_REPLACE, 'nsslapd-security', [b'on'])]
        (ret, return_value) = ldap_obj.modify_ldap(mod_dn3, enable_security)
        if not return_value:
            raise LdapException(
                'Failed to enable nsslapd-security, Error:%s' % (ret))
        else:
            print('Enabled nsslapd-security')

        # set the appropriate TLS port
        mod_dn4 = 'cn=config'
        enable_ssl_port = [(ldap.MOD_REPLACE, 'nsslapd-securePort',
                            str(tls_port).encode())]
        (ret, return_value) = ldap_obj.modify_ldap(mod_dn4, enable_ssl_port)
        if not return_value:
            raise LdapException(
                'Failed to set nsslapd-securePort, Error:%s' % (ret))
        else:
            print('Enabled nsslapd-securePort=%r' % tls_port)


class DirSrvWrap(object):
    """This is a wrapper class for DirSrv.

    This is a wrapper class of DirSrv class which validates
    all the inputs sent to Dirsrv object. Specifies ports for
    LDAP and TLS ports, specifies default suffix.
    """
    # pylint: disable=too-many-instance-attributes
    def __init__(self, multihost_obj, client_obj=None,
                 ssl=None, ssldb=None, canick=None):
        """
        Create a DirSrv object for a specific Host. Specify the ports,
        instance details to the Dirsrv object

        Args:
            multihost_obj (obj): Multihost object
            ssl (bool): set True to enable SSL else none/False
            ssldb (str): Directory containing CA and server certs
        """
        self.ds_used_ports = {}
        self.dirsrv_info = {}
        self.dirsrv_obj = None
        self.ds_instance_name = None
        self.multihost = multihost_obj
        self.ds_instance_host = self.multihost.sys_hostname
        self.client_host = client_obj
        self.ds_instance_suffix = None
        self.ds_rootdn_pwd = None
        self.ds_ldap_port = None
        self.ds_tls_port = None
        self.ssl = ssl
        if self.ssl:
            self.ssl_dir = ssldb
        self.canick = canick

    def __iter__(self):
        """ iter values of each instance """
        return self.dirsrv_info.itervalues()

    def __getitem__(self, key):
        """ Return values of each instance """
        return self.dirsrv_info[key]

    def _set_options(self,
                     instance_name,
                     instance_suffix,
                     root_dn_pwd,
                     ldap_port,
                     tls_port):
        """set Default values.

        Args:
            instance_name (str): DS Instance Name
            instance_suffix (str): DS Instance Suffix
            root_dn_pwd (str): Directory Manager password
            ldap_port (str): LDAP port
            tls_port (str): TLS Port

        Returns:
              A tuple containing 'Success', 0 or Failed Message and 1
              for failure to setup ports

        Exceptions:
              None
        """
        self.ds_instance_name = instance_name
        if instance_suffix:
            self.ds_instance_suffix = instance_suffix
        else:
            self.ds_instance_suffix = 'dc=example,dc=org'

        if root_dn_pwd:
            self.ds_rootdn_pwd = root_dn_pwd
        else:
            self.ds_rootdn_pwd = 'Secret123'
        # Get ports
        try:
            self.ds_ldap_port, self.ds_tls_port = self._set_ports(
                ldap_port, tls_port)
        except IndexError:
            return "No more ports available", 1
        else:
            self.ds_used_ports[self.ds_instance_name] = [
                self.ds_ldap_port, self.ds_tls_port]

        # validate instance
        try:
            self._validate_options()
        except DirSrvException as err:
            return err.msg, err.rval
        else:
            return "Success", 0

    def _set_ports(self, u_port, e_port):
        """return ports required to setup DS Instance.

        Idea behind this is when a directory server instance needs
        to be created we need ports for LDAP and SSL ports.
        1. check if LDAP port and SSL port is given
        1.1 If given, verify if the ports are available(not used)
        1.1.1 Bind that port to ldap_port_t using semanage command
        1.1.2 Use the ports and add it to the self.UsedPorts list
        1.2 else raise exception
        2. If LDAP port and SSL port is not given
        2.1 Check if the ports are available(not used)
        2.1.1 Bind the port to ldap_port_t using semanage command
        2.1.2 Use the ports and add it to self.ds_used_ports list

        Args:
           u_port (str): unencrypted LDAP port
           e_port (str): Encrypted port to be used for TLS

        Returns:
              sorted_available_ports[0] (list): LDAP and TLS ports
        """
        ds_ports = [389, 1389, 2389, 3389, 4389, 30389, 31389, 32389, 33389,
                    34389, 35389, 36389, 37389, 38389, 39389]
        tls_ports = [636, 1636, 2636, 3636, 4636, 30636, 31636, 32636, 33636,
                     34636, 35636, 36636, 37636, 38636, 39636]

        if u_port is None and e_port is None:
            for ldap_port, ldaps_port in zip(ds_ports, tls_ports):
                if (self._check_remote_port(ldap_port)
                        or self._check_remote_port(ldaps_port)):
                    pass
                else:
                    return ldap_port, ldaps_port
        else:
            list_a = []
            for ports in self.ds_used_ports.values():
                list_a.append(ports)

            list_b = []
            for l_port, s_port in zip(ds_ports, tls_ports):
                list_b.append((l_port, s_port))

            if len(set(list_a)) > len(set(list_b)):
                available_ports = set(list_a) - set(list_b)
            else:
                available_ports = set(list_b) - set(list_a)
            sorted_available_ports = sorted(available_ports)
            return sorted_available_ports[0]

    def _check_remote_port(self, port):
        """check if the port on the remote host is free.

        Args:
            port (int): check if port is available

        Returns:
            bool: True if port is free else False.
        """
        sock_obj = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock_obj.settimeout(1)
        try:
            sock_obj.connect((self.ds_instance_host, port))
        except socket.error as err:
            print("fail to connect to port %s due to error %r" % (port,
                                                                  err.errno))
            return False
        else:
            sock_obj.close()
            return True

    def _validate_options(self):
        """verify if the instance directory already exists.

        Args:
            None

        Returns:
            None

        Exceptions:
            DirSrvException: if instance directory already exists
        """
        check_instance = ['/usr/bin/ls', '/etc/dirsrv/slapd-%s' %
                          self.ds_instance_name]
        output = self.multihost.run_command(check_instance, raiseonerr=False)
        if output.returncode == 0:
            raise DirSrvException('%s Instance already exists' %
                                  self.ds_instance_name)

    def create_ds_instance(self,
                           inst_name,
                           inst_suffix=None,
                           root_dn_pwd=None,
                           ldap_port=None,
                           tls_port=None):
        """Create Directory Server Instance.

        Args:
            inst_name (str): Instance Name
            inst_suffix (str): Instance suffix
            root_dn_pwd (str): Directory Manager password
            ldap_port (str): LDAP port to be used
            tls_port (str): TLS port to be used

        Returns:
            result (str) and return code (str): Result containing message and
            return code containing 0 or 1 (1 indicating failure)

        Exceptions:
            DirSrvException: if DS instance could not be created.
        """
        result, return_code = self._set_options(inst_name,
                                                inst_suffix,
                                                root_dn_pwd,
                                                ldap_port,
                                                tls_port)
        if return_code == 0:
            self.dirsrv_obj = DirSrv(name=self.ds_instance_name,
                                     host=self.ds_instance_host,
                                     suffix=self.ds_instance_suffix,
                                     multihost=self.multihost,
                                     root_dn_password=self.ds_rootdn_pwd,
                                     ldap_port=self.ds_ldap_port,
                                     tls_port=self.ds_tls_port)
            cfg_file = self.dirsrv_obj.create_config()
            try:
                self.dirsrv_obj.setup_ds(cfg_file)
            except subprocess.CalledProcessError:
                raise DirSrvException('Failed to setup Directory server')
            self.dirsrv_info[self.ds_instance_name] = self.dirsrv_obj.__dict__
            ldap_uri = 'ldap://%s:%r' % (self.ds_instance_host,
                                         self.ds_ldap_port)
            try:
                self.dirsrv_obj.enable_anonymous_search(ldap_uri)
            except LdapException:
                raise DirSrvException("Failed to enable anonymous search")
            if self.ssl:
                try:
                    self.dirsrv_obj.setup_certs(self.ssl_dir,
                                                self.client_host, self.canick)
                except DirSrvException as err:
                    return err.msg, err.rval
                else:
                    (result, return_code) = self.enablessl()
            return result, return_code
        else:
            raise DirSrvException('Failed to setup Directory Server instance')

    def enablessl(self):
        """Enable SSL/TLS on instance.

        Enable by adding TLS port to ldap_port_t SELinux label and restart
        Directory Server.

        Args:
            None

        Returns:
            Tuple: Success, 0 or Error, 1

        Exceptions:
            None:
        """
        # add TLS port to ldap_port_t SELinux label

        add_tls_port = ['semanage', 'port', '-a', '-t',
                        'ldap_port_t', '-p', 'tcp', str(self.ds_tls_port)]

        restart_ds = ['systemctl', 'restart', 'dirsrv@%s' %
                      self.ds_instance_name]
        if self.ds_tls_port != 636:
            try:
                self.multihost.run_command(add_tls_port)
            except subprocess.CalledProcessError:
                return "Failed to set tls_port as ldap_port_t", 1
            else:
                self.multihost.log.info('Added %s port to ldap_port_t' %
                                        self.ds_tls_port)
        try:
            self.dirsrv_obj.enable_ssl('ldap://%s:%r' % (self.ds_instance_host,
                                                         self.ds_ldap_port),
                                       self.ds_tls_port)
        except LdapException:
            return "Error", 1

        try:
            self.multihost.run_command(restart_ds)
        except subprocess.CalledProcessError:
            return "Error", 1
        else:
            self.multihost.log.info('DS instance restarted successfully')
            # sleep for 10 seconds
            time.sleep(10)
            tail_cmd = ['tail', '-n', '100',
                        '/var/log/dirsrv/slapd-%s/errors' % self.
                        ds_instance_name]
            output = self.multihost.run_command(tail_cmd, raiseonerr=False)
            if output.returncode != 0:
                return "Error", 1
            else:
                return "Success", 0

    def remove_ds_instance(self, instance_name):
        """remove Directory server instance.

        Args:
            instance_name (str): Instance Name

        Returns:
            bool: True if successfully removed

        Exceptions:
            DirSrvException: if DS instance cannot be removed
        """
        ret = self.dirsrv_info[instance_name]
        if ret['instance_name'] == instance_name:
            inst_name = ret['ds_inst_name']
            try:
                self.dirsrv_obj.remove_ds(inst_name)
            except subprocess.CalledProcessError:
                raise DirSrvException('Failed to remove %s instance',
                                      inst_name)
            else:
                del self.ds_used_ports[instance_name]
                return True
        else:
            raise DirSrvException('%s Instance not found' % instance_name)



"""
Access Control - Group Policy Objects (GPO) Tests

Switch user (su), and Remote (ssh) access is parameterized fixture for the tests. When SeRemoteInteractiveLogonRight is
omitted from the policy, which are most tests cases, the value from SeInteractiveLogonRight is then copied to
SeRemoteInteractiveLogonRight.

The following code will modify both SeInteractiveActiveLogonRight and SeRemoteInteractiveLogonRight.

.. code-block::
    provider.gpo("site policy").add().policy(
        {
            "SeInteractiveLogonRight": [user1, group, provider.group("Domain Admins")],
            "SeDenyInteractiveLogonRight": [],
        }
       )

An administrative user or group always needs to be specified, to prevent administrative lock outs, for the tests
"Domain Admins" group is used.

The following GPO related parameters are not tested
- ad_gpo_cache_timeout
- ad_gpo_map_network
- ad_gpo_map_batch
- ad_gpo_map_service
- ad_gpo_map_permit
- ad_gpo_map_deny


:requirement: Access Control
"""

from __future__ import annotations

import pytest
from sssd_test_framework.roles.ad import AD
from sssd_test_framework.roles.client import Client
from sssd_test_framework.roles.generic import GenericADProvider
from sssd_test_framework.roles.samba import Samba
from sssd_test_framework.topology import KnownTopology, KnownTopologyGroup


@pytest.mark.importance("critical")
@pytest.mark.parametrize("method", ["su", "ssh"])
@pytest.mark.topology(KnownTopologyGroup.AnyAD)
def test_gpo__is_set_to_enforcing(client: Client, provider: GenericADProvider, method: str):
    """
    :title: Group policy object host base access control is set to enforcing and users are allowed
    :description:
        When ad_gpo_access_control is set to enforcing, user access is managed and enforced by the policy.
        Unlisted users are denied access.
    :setup:
        1. Create the following users; 'user', 'user1', 'user2', 'deny_user1', 'deny_user2'
        2. Create the following groups; 'group' with member 'user2', 'deny_group' with member 'deny_user2'
        3. Create and link the GPO 'site policy' and add 'user1', 'group' and 'Domain Admins' to
           SeInteractiveLogonRight key and 'deny_user1' and 'deny_group' to SeDenyInteractiveLogonRight key
        4. Configure sssd.conf with 'ad_gpo_access_control' = 'enforcing'
        5. Start SSSD
    :steps:
        1. Authenticate as 'user'
        2. Authenticate as 'user1' and 'user2'
        3. Authenticate as 'deny_user1' and 'deny_user2'
    :expectedresults:
        1. User authentication is unsuccessful
        2. User authentications are successful
        3. User authentications are unsuccessful
    :customerscenario: True
    """
    provider.user("user").add()
    user1 = provider.user("user1").add()
    user2 = provider.user("user2").add()
    deny_user1 = provider.user("deny_user1").add()
    deny_user2 = provider.user("deny_user2").add()
    group = provider.group("group").add().add_members([user2])
    deny_group = provider.group("deny_group").add().add_members([deny_user2])

    provider.gpo("site policy").add().policy(
        {
            "SeInteractiveLogonRight": [user1, group, provider.group("Domain Admins")],
            "SeDenyInteractiveLogonRight": [deny_user1, deny_group],
        }
    ).link()

    client.sssd.domain["ad_gpo_access_control"] = "enforcing"
    client.sssd.start()

    assert not client.auth.parametrize(method).password(
        username="user", password="Secret123"
    ), "User absent from policy authenticated successfully!"

    assert client.auth.parametrize(method).password(
        username="user1", password="Secret123"
    ), "Allowed user authentication failed!"

    assert client.auth.parametrize(method).password(
        username="user2", password="Secret123"
    ), "Allowed group user authentication failed!"

    assert not client.auth.parametrize(method).password(
        username="deny_user1", password="Secret123"
    ), "Denied user authenticated successfully!"

    assert not client.auth.parametrize(method).password(
        username="deny_user2", password="Secret123"
    ), "Denied group user authenticated successfully!"


@pytest.mark.importance("critical")
@pytest.mark.parametrize("method", ["su", "ssh"])
@pytest.mark.topology(KnownTopologyGroup.AnyAD)
def test_gpo__is_set_to_enforcing_with_no_policy(client: Client, provider: GenericADProvider, method: str):
    """
    :title: Group policy object host base access control is set to enforcing with no policy
    :description:
        When ad_gpo_access_control is set to enforcing with no policy applied. It should
        permit all users to login.
    :setup:
        1. Create the following user, 'user'
        2. Configure sssd.conf with 'ad_gpo_access_control' = 'enforcing'
        3. Start SSSD
    :steps:
        1. Authenticate 'user'
        2. Check logs for gpo_access_check result
    :expectedresults:
        1. User authentication is successful
        2. Access check result is granted
    :customerscenario: True
    """
    provider.user("user").add()

    client.sssd.domain["ad_gpo_access_control"] = "enforcing"
    client.sssd.start()

    assert client.auth.parametrize(method).password(username="user", password="Secret123")


@pytest.mark.importance("critical")
@pytest.mark.parametrize("method", ["su", "ssh"])
@pytest.mark.topology(KnownTopologyGroup.AnyAD)
def test_gpo__is_set_to_permissive_and_users_are_allowed(client: Client, provider: GenericADProvider, method: str):
    """
    :title: Group policy object host base access control is set to permissive
    :description:
        When ad_gpo_access_control is set to permissive, all users are able to login despite what
        is configured in the policy. The rules are still processed and can be viewed in the logs.
    :setup:
        1. Create the following user, 'user1'
        2. Create and link the GPO 'site policy' and add 'user1', group and 'Domain Admins' to
           SeInteractiveLogonRight key.
        3. Configure sssd.conf with 'ad_gpo_access_control' = 'permissive'
        4. Start SSSD
    :steps:
        1. Authenticate 'user1'
        2. Check logs for ad_gpo_access_control value
        3. Check logs for gpo_access_check result
    :expectedresults:
        1. User authentication is successful
        2. ad_gpo_access_control is permissive
        3. Access check result is granted
    :customerscenario: True
    """
    user1 = provider.user("user1").add()

    provider.gpo("site policy").add().policy(
        {
            "SeInteractiveLogonRight": [user1, provider.group("Domain Admins")],
            "SeDenyInteractiveLogonRight": [],
        }
    ).link()

    client.sssd.domain["ad_gpo_access_control"] = "permissive"
    client.sssd.start()

    assert client.auth.parametrize(method).password(
        username="user1", password="Secret123"
    ), "(Permissive) Allowed user authentication failed!"
    log_str = client.fs.read(client.sssd.logs.domain())

    assert (
        "Option ad_gpo_access_control has value permissive" in log_str
    ), "'Option ad_gpo_access_control has value permissive' not in logs!"

    assert "access_granted = 1" in log_str, "'access_granted = 1' not in logs!"


@pytest.mark.importance("critical")
@pytest.mark.parametrize("method", ["su", "ssh"])
@pytest.mark.topology(KnownTopologyGroup.AnyAD)
def test_gpo__is_set_to_permissive_and_users_are_denied(client: Client, provider: GenericADProvider, method: str):
    """
    :title: Group policy object host base access control is set to permissive
    :description:
        When ad_gpo_access_control is set to permissive, all users are able to login despite what
        is configured in the policy. The rules are still processed and can be viewed in the logs.
    :setup:
        1. Create the following user, 'deny_user1'
        2. Create and link the GPO 'site policy' and add group 'Domain Admins' to
           SeInteractiveLogonRight key. Add 'deny_user1' to SeDenyInteractiveLogonRight key
        3. Configure sssd.conf with 'ad_gpo_access_control' = 'permissive'
        4. Start SSSD
    :steps:
        1. Authenticate 'deny_user1'
        2. Check logs for ad_gpo_access_control value
        3. Check logs for gpo_access_check result
    :expectedresults:
        1. User authentication is successful
        2. ad_gpo_access_control is permissive
        3. Access check result is denied
    :customerscenario: True
    """
    deny_user1 = provider.user("deny_user1").add()

    provider.gpo("site policy").add().policy(
        {
            "SeInteractiveLogonRight": [provider.group("Domain Admins")],
            "SeDenyInteractiveLogonRight": [deny_user1],
        }
    ).link()

    client.sssd.domain["ad_gpo_access_control"] = "permissive"
    client.sssd.start()

    assert client.auth.parametrize(method).password(
        username="deny_user1", password="Secret123"
    ), "(Permissive) Denied user authentication failed!"

    log_str = client.fs.read(client.sssd.logs.domain())
    assert (
        "Option ad_gpo_access_control has value permissive" in log_str
    ), "'Option ad_gpo_access_control has value permissive' not in logs!"

    assert "access_granted = 0" in log_str, "'access_granted = 1' not in logs!"


@pytest.mark.importance("critical")
@pytest.mark.parametrize("method", ["su", "ssh"])
@pytest.mark.topology(KnownTopologyGroup.AnyAD)
def test_gpo__is_set_to_disabled_and_all_users_are_allowed(client: Client, provider: GenericADProvider, method: str):
    """
    :title: Group policy object host base access control is set to disabled and all users are allowed
    :description:
        When ad_gpo_access_control is set to disabled, no policy processing should occur and be
        turned off entirely.
    :setup:
        1. Create the following users; 'user', 'user1' and 'deny_user1'
        2. Create and link the GPO 'site policy' and add 'user1', group and 'Domain Admins' to
           SeInteractiveLogonRight key, and add 'deny_user1' to SeDenyInteractiveLogonRight key
        3. Configure sssd.conf with 'ad_gpo_access_control' = 'disabled'
        4. Start SSSD
    :steps:
        1. Authenticate as 'user', 'user1' and 'deny_user1'
        2. Check logs for ad_gpo_access_control value
    :expectedresults:
        1. All users authentication are successful
        2. ad_gpo_access_control is disabled
    :customerscenario: True
    """
    provider.user("user").add()
    user1 = provider.user("user1").add()
    deny_user1 = provider.user("deny_user1").add()

    provider.gpo("site policy").add().policy(
        {
            "SeInteractiveLogonRight": [user1, provider.group("Domain Admins")],
            "SeDenyInteractiveLogonRight": [deny_user1],
        }
    ).link()

    client.sssd.domain["ad_gpo_access_control"] = "disabled"
    client.sssd.start()

    assert client.auth.parametrize(method).password(
        username="user", password="Secret123"
    ), "(Disabled) User authentication failed!"

    assert client.auth.parametrize(method).password(
        username="user1", password="Secret123"
    ), "(Disabled) User authentication failed!"

    assert client.auth.parametrize(method).password(
        username="deny_user1", password="Secret123"
    ), "(Disabled) User authentication failed!"

    log_str = client.fs.read(client.sssd.logs.domain())
    assert (
        "Option ad_gpo_access_control has value disabled" in log_str
    ), "'Option ad_gpo_access_control has value disabled' not in logs!"


@pytest.mark.importance("critical")
@pytest.mark.parametrize("method", ["ssh", "su"])
@pytest.mark.topology(KnownTopologyGroup.AnyAD)
@pytest.mark.ticket(bz=1695576)
def test_gpo__implicit_deny_is_set_to_true(client: Client, provider: GenericADProvider, method: str):
    """
    :title: Group policy object host base access control is set to enforcing and implicit deny is true
    :description:
        When ad_gpo_implicit_deny is set to true, the new default behavior is to deny access for
        when there is no policy applied.
    :setup:
        1. Create the following users; 'user'
        2. Configure sssd.conf with 'ad_gpo_access_control' = 'enforcing' and 'ad_gpo_implicit_deny' = True
        3. Start SSSD
    :steps:
        1. Authenticate as 'user'
    :expectedresults:
        1. 'user' authentication is unsuccessful
    :customerscenario: True
    """
    provider.user("user").add()

    client.sssd.domain["ad_gpo_access_control"] = "enforcing"
    client.sssd.domain["ad_gpo_implicit_deny"] = "True"
    client.sssd.start()

    assert not client.auth.parametrize(method).password(
        username="user", password="Secret123"
    ), "User authenticated successfully!"


@pytest.mark.importance("critical")
@pytest.mark.parametrize("method", ["ssh", "su"])
@pytest.mark.topology(KnownTopologyGroup.AnyAD)
def test_gpo__domain_and_sites_inheritance_when_site_is_enforcing(
    client: Client, provider: GenericADProvider, method: str
):
    """
    :title: Group policy object host base access control checking inheritance for sites enforced and domains
    :description:
        Policies can be applied to three locations, sites, domains and OUs. The order of which
        these are processed are, OUs, domains and lastly sites. Policies can be set to be enforced,
        which puts the policy at the top. The keys are not constructed and is replaced with the value of
        the policy with the highest priority.
    :setup:
        1. Create the following users; user1, user2
        2. Create and link the 'site policy' and add 'user1' and 'Domain Admins' to
           SeInteractiveLogonRight and add 'user2' to SeDenyInteractiveLogonRight key
        3. Create and link the 'domain policy' and add 'user2' and 'Domain Admins' to
           SeInteractiveLogonRight key. Add 'user1' to SeDenyInteractiveLogonRight key
        4. Set 'site policy' to be enforcing
        5. Configure sssd.conf with 'ad_gpo_access_control' = 'enforcing'
        6. Start SSSD
    :steps:
        1. Authenticate as 'user1'
        2. Authenticate as 'user2'
    :expectedresults:
        1. 'user1' authentication is successful
        2. 'user2' authentication is unsuccessful
    :customerscenario: True
    """
    user1 = provider.user("user1").add()
    user2 = provider.user("user2").add()

    provider.gpo("site policy").add().policy(
        {
            "SeInteractiveLogonRight": [user1, provider.group("Domain Admins")],
            "SeDenyInteractiveLogonRight": [user2],
        }
    ).link(enforced=True)

    provider.gpo("domain policy").add().policy(
        {
            "SeInteractiveLogonRight": [user2, provider.group("Domain Admins")],
            "SeDenyInteractiveLogonRight": [user1],
        }
    ).link(target=f"{provider.naming_context}")

    client.sssd.domain["ad_gpo_access_control"] = "enforcing"
    client.sssd.start()

    assert client.auth.parametrize(method).password(
        username="user1", password="Secret123"
    ), "User authentication failed!"

    assert not client.auth.parametrize(method).password(
        username="user2", password="Secret123"
    ), "User authenticated successfully!"


@pytest.mark.importance("critical")
@pytest.mark.parametrize("method", ["ssh", "su"])
@pytest.mark.topology(KnownTopologyGroup.AnyAD)
def test_gpo__domain_and_sites_inheritance(client: Client, provider: GenericADProvider, method: str):
    """
    :title: Group policy object host base access control checking inheritance for sites and domains.
    :description:
        Policies can be applied to three locations, sites, domains and OUs. The order of which
        these are processed are, OUs, domains and lastly sites. Policies can be set to be enforced,
        which puts the policy at the top. The keys are not constructed and is replaced with the value of
        the policy with the highest priority.
    :setup:
        1. Create the following users; user1, user2
        2. Create and link the 'site policy' and add 'user1' and 'Domain Admins' to
           SeInteractiveLogonRight and add 'user2' to SeDenyInteractiveLogonRight key
        3. Create and link the 'domain policy' and add 'user2' and 'Domain Admins' to
           SeInteractiveLogonRight key. Add 'user3' to SeDenyInteractiveLogonRight key
        4. Configure sssd.conf with 'ad_gpo_access_control' = 'enforcing'
        5. Start SSSD
    :steps:
        1. Authenticate as 'user1'
        2. Authenticate as 'user2'
    :expectedresults:
        1. 'user1' authentication is unsuccessful
        2. 'user2' authentication is successful
    :customerscenario: True
    """
    user1 = provider.user("user1").add()
    user2 = provider.user("user2").add()

    provider.gpo("site policy").add().policy(
        {
            "SeInteractiveLogonRight": [user1, provider.group("Domain Admins")],
            "SeDenyInteractiveLogonRight": [user2],
        }
    ).link()

    provider.gpo("domain policy").add().policy(
        {
            "SeInteractiveLogonRight": [user2, provider.group("Domain Admins")],
            "SeDenyInteractiveLogonRight": [user1],
        }
    ).link(target=f"{provider.naming_context}")

    client.sssd.domain["ad_gpo_access_control"] = "enforcing"
    client.sssd.start()

    assert not client.auth.parametrize(method).password(
        username="user1", password="Secret123"
    ), "Site user authenticated successfully!"

    assert client.auth.parametrize(method).password(
        username="user2", password="Secret123"
    ), "Domain user authentication failed!"


@pytest.mark.importance("critical")
@pytest.mark.parametrize("method", ["ssh", "su"])
@pytest.mark.topology(KnownTopologyGroup.AnyAD)
def test_gpo__ou_and_domain_inheritance(client: Client, provider: AD, method: str):
    """
    :title: Group policy object host base access control checking inheritance between ous and domains.
    :description:
        Policies can be applied to three locations, sites, domains and OUs. The order of which
        these are processed are, OUs, domains and lastly sites. Policies can be set to be enforced,
        which puts the policy at the top. The keys are not constructed and is replaced with the value of
        the policy with the highest priority.
    :setup:
        1. Create the following users; user1, user2
        2. Create test OU
        3. Create and link the 'domain policy' and add 'user1' and 'Domain Admins' to
           SeInteractiveLogonRight key. Add 'user2' to SeDenyInteractiveLogonRight key
        4. Create and link the 'ou policy' and add 'user2' and 'Domain Admins' to
           SeInteractiveLogonRight key and add 'user2' to SeDenyInteractiveLogonRight key
        5. Move computer object to test OU
        6. Configure sssd.conf with 'ad_gpo_access_control' = 'enforcing'
        7. Start SSSD
    :steps:
        1. Authenticate as 'user1'
        2. Authenticate as 'user2'
    :expectedresults:
        1. 'user1' authentication is unsuccessful
        2. 'user2' authentication is successful
    :customerscenario: True
    """
    user1 = provider.user("user1").add()
    user2 = provider.user("user2").add()
    ou = provider.ou("test").add().dn

    provider.gpo("domain policy").add().policy(
        {
            "SeInteractiveLogonRight": [user1, provider.group("Domain Admins")],
            "SeDenyInteractiveLogonRight": [user2],
        }
    ).link(target=f"{provider.host.naming_context}")

    provider.gpo("ou policy").add().policy(
        {
            "SeInteractiveLogonRight": [user2, provider.group("Domain Admins")],
            "SeDenyInteractiveLogonRight": [user1],
        }
    ).link(target=ou)

    provider.computer(client.host.hostname.split(".")[0]).move(ou)

    client.sssd.domain["ad_gpo_access_control"] = "enforcing"
    client.sssd.start()

    assert not client.auth.parametrize(method).password(
        username="user1", password="Secret123"
    ), "Domain user authenticated successfully!"

    assert client.auth.parametrize(method).password(
        username="user2", password="Secret123"
    ), "OU user authentication failed!"


@pytest.mark.importance("critical")
@pytest.mark.parametrize("method", ["ssh", "su"])
@pytest.mark.topology(KnownTopology.AD)
@pytest.mark.ticket(bz=1804005)
def test_gpo__sites_inheritance_using_gpo_link_order(client: Client, ad: AD, method: str):
    """
    :title: Group policy object host base access control checking inheritance using GPO link order.
    :description:
        When a policy is applied to the 'same target', there is an additional level of inheritance, the
        order. This is automatically created as policies are linked ot the target and can be changed.
    :setup:
        1. Create the following users; user, user1, user2
        2. Create and link the 'site policy 1' and add 'user1' and 'Domain Admins' to
           SeInteractiveLogonRight and add 'user2' to SeDenyInteractiveLogonRight key
        3. Create and link the 'site policy 2' and add 'user2' and 'Domain Admins' to
           SeInteractiveLogonRight key. Add 'user1' to SeDenyInteractiveLogonRight key with link order 1
        4. Configure sssd.conf with 'ad_gpo_access_control' = 'enforcing'
        5. Start SSSD
    :steps:
        1. Authenticate as 'user1'
        2. Authenticate as 'user2'
    :expectedresults:
        1. 'user1' authentication is unsuccessful
        2. 'user2' authentication is successful
    :customerscenario: True
    """
    user1 = ad.user("user1").add()
    user2 = ad.user("user2").add()

    ad.gpo("site policy 1").add().policy(
        {
            "SeInteractiveLogonRight": [user1, ad.group("Domain Admins")],
            "SeDenyInteractiveLogonRight": [user2],
        }
    ).link()

    ad.gpo("site policy 2").add().policy(
        {
            "SeInteractiveLogonRight": [user2, ad.group("Domain Admins")],
            "SeDenyInteractiveLogonRight": [user1],
        }
    ).link(order=1)

    client.sssd.domain["ad_gpo_access_control"] = "enforcing"
    client.sssd.start()

    assert not client.auth.parametrize(method).password(
        username="user1", password="Secret123"
    ), "User authenticated successfully!"

    assert client.auth.parametrize(method).password(username="user2", password="Secret123"), "User failed login!"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopologyGroup.AnyAD)
def test_gpo__map_interactive_disabling_login_su_and_su_l(client: Client, provider: GenericADProvider):
    """
    :title: Group policy object host based access disabling logon, su, su-l GPO evaluation.
    :description:
        Setting ad_gpo_map_interactive, changes the evaluation of the PAM services specified, by default,
        this setting changes evaluates the following PAM services, login, su, su-l, gdm-fingerprint,
        gdm-password, gdm-smartcard, kdm, lightdm, lxdm, sddm, unity and xdm
    :setup:
        1. Create the following users; 'user1', 'deny_user1'
        2. Create and link the GPO 'site policy' and add 'user1', group and 'Domain Admins' to
           SeInteractiveLogonRight key. Add 'deny_user1' SeDenyInteractiveLogonRight key
        3. Configure sssd.conf with 'ad_gpo_access_control' = 'enforcing' and
           'ad_gpo_map_interactive' = "-logon -su -su-l"
        4. Start SSSD
    :steps:
        1. Authenticate as 'user1' with su
        2. Authenticate as 'user1' with ssh
        3. Authenticate as 'deny_user1' with su
        4. Authenticate as 'deny_user1' with ssh
    :expectedresults:
        1. 'user1' authentication is unsuccessful
        2. 'user1' authentication is successful
        3. 'deny_user1' authentication is unsuccessful for su
        4. 'deny_user1' authentication is unsuccessful for ssh
    :customerscenario: True
    """
    user1 = provider.user("user1").add()
    deny_user1 = provider.user("deny_user1").add()

    provider.gpo("site policy").add().policy(
        {
            "SeInteractiveLogonRight": [user1, provider.group("Domain Admins")],
            "SeDenyInteractiveLogonRight": [deny_user1],
        }
    ).link()

    client.sssd.domain["ad_gpo_access_control"] = "enforcing"
    client.sssd.domain["ad_gpo_map_interactive"] = "-logon, -su, -su-l"
    client.sssd.start()

    assert not client.auth.su.password("user1", password="Secret123"), "Allowed user, authenticated SU successfully!"
    assert client.auth.ssh.password("user1", password="Secret123"), "Allowed user SSH authentication failed!"

    assert not client.auth.su.password(
        "deny_user1", password="Secret123"
    ), "Denied user, authenticated SU successfully!"

    assert not client.auth.ssh.password(
        "deny_user1", password="Secret123"
    ), "Denied user authenticated SSH successfully!"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopologyGroup.AnyAD)
def test_gpo__map_remote_interactive_disabling_sshd(client: Client, provider: GenericADProvider):
    """
    :title: Group policy object host based access disabling ssh and cockpit GPO evaluation.
    :description:
        Setting ad_gpo_map_remote_interactive, changes the evaluation of the PAM services specified, by default,
        this setting changes evaluates the following PAM services, sshd, cockpit
    :setup:
        1. Create the following users; 'user1', 'deny_user1'
        2. Create and link the GPO 'site policy' and add 'user1', group and 'Domain Admins' to
           SeInteractiveLogonRight key. Add 'deny_user1' SeDenyInteractiveLogonRight key
        3. Configure sssd.conf with 'ad_gpo_access_control' = 'enforcing' and
           'ad_gpo_map_remote_interactive' = '-sshd'
        4. Start SSSD
    :steps:
        1. Authenticate 'user1' with su
        2. Authenticate 'user1' with ssh
        3. Authenticate 'deny_user1' with su
        4. Authenticate 'deny_user1' with ssh
    :expectedresults:
        1. 'user1' authentication is successful
        2. 'user1' authentication is unsuccessful
        3. 'deny_user1' authentication is unsuccessful for su
        4. 'deny_user1' authentication is unsuccessful for ssh
    :customerscenario: True
    """
    user1 = provider.user("user1").add()
    deny_user1 = provider.user("deny_user1").add()

    provider.gpo("site policy").add().policy(
        {
            "SeInteractiveLogonRight": [user1, provider.group("Domain Admins")],
            "SeDenyInteractiveLogonRight": [deny_user1],
        }
    ).link()

    client.sssd.domain["ad_gpo_access_control"] = "enforcing"
    client.sssd.domain["ad_gpo_map_remote_interactive"] = "-sshd"
    client.sssd.start()

    assert client.auth.su.password("user1", password="Secret123"), "Allowed user SU authentication failed!"
    assert not client.auth.ssh.password("user1", password="Secret123"), "Allowed user, authenticated SSH successfully!"

    assert not client.auth.su.password(
        "deny_user1", password="Secret123"
    ), "Denied user, authenticated SU successfully!"

    assert not client.auth.ssh.password(
        "deny_user1", password="Secret123"
    ), "Denied user, authenticated SSH successfully!"


@pytest.mark.importance("critical")
@pytest.mark.parametrize("method", ["ssh", "su"])
@pytest.mark.topology(KnownTopologyGroup.AnyAD)
def test_gpo__works_when_the_server_is_unreachable(client: Client, provider: GenericADProvider, method: str):
    """
    :title: Group policy object host based works when the server is unreachable.
    :description: Tests that gpo processing works from the cache when the server is unreachable
    :setup:
        1. Create the following users; 'user1', 'deny_user1'
        2. Create and link the GPO 'site policy' and add 'user1', group and 'Domain Admins' to
           SeInteractiveLogonRight key. Add 'deny_user1' to SeDenyInteractiveLogonRight key
        3. Configure sssd.conf with 'ad_gpo_access_control' = 'enforcing', 'cache_credentials' = 'True'
           'krb5_store_password_if_offline' = 'True' and 'offline_credentials_expiration' = '0'
        4. Start SSSD
    :steps:
        1. Authenticate 'user1'
        2. Authenticate 'deny_user1'
        3. Use iptables and block traffic to the ad server
        4. Authenticate 'user1'
        5. Authenticate 'deny_user1'
    :expectedresults:
        1. 'user1' authentication is successful
        2. 'deny_user1' authentication is unsuccessful
        3. Traffic is blocked and SSSD goes into offline mode
        4. 'user1' authentication is successful
        5. 'deny_user1' authentication is unsuccessful
    :customerscenario: True
    """
    user1 = provider.user("user1").add()
    deny_user1 = provider.user("deny_user1").add()

    provider.gpo("site policy").add().policy(
        {
            "SeInteractiveLogonRight": [user1, provider.group("Domain Admins")],
            "SeDenyInteractiveLogonRight": [deny_user1],
        }
    ).link()

    client.sssd.domain["ad_gpo_access_control"] = "enforcing"
    client.sssd.domain["cache_credentials"] = "True"
    client.sssd.domain["krb5_store_password_if_offline"] = "True"
    client.sssd.pam["offline_credentials_expiration"] = "0"
    client.sssd.start()

    assert client.auth.parametrize(method).password(
        "user1", password="Secret123"
    ), "Allowed user authentication failed!"

    assert not client.auth.parametrize(method).password(
        "deny_user1", password="Secret123"
    ), "Denied user authenticated successfully!"

    client.firewall.outbound.drop_host(provider)
    client.sssd.bring_offline()

    assert client.auth.parametrize(method).password(
        "user1", password="Secret123"
    ), "Allowed user authentication failed!"

    assert not client.auth.parametrize(method).password(
        "deny_user1", password="Secret123"
    ), "Denied user authenticated successfully!"


@pytest.mark.importance("critical")
@pytest.mark.parametrize("method", ["ssh", "su"])
@pytest.mark.topology(KnownTopologyGroup.AnyAD)
@pytest.mark.ticket(bz=1547234)
def test_gpo__honors_the_ad_site_parameter(client: Client, provider: GenericADProvider, method: str):
    """
    :title: Group policy object host based access control honors the ad_site parameter in the configuration.
    :description:
        The ad_site parameter use to break processing the GPO feature entirely. This ensures that setting
        the parameter doesn't break the feature. Also the site is auto-discovered, and the default site is
        'Default-First-Site-Name', when specifying a site

        Note: The site name cannot contain spaces
    :setup:
        1. Create the following users; 'user1', 'deny_user1'
        2. Create new site
        3. Create the GPO 'site policy' and add 'user1', group and 'Domain Admins' to
           SeInteractiveLogonRight key. Add 'deny_user1' to SeDenyInteractiveLogonRight key
           and link policy to the new site
        4. Configure sssd.conf with 'ad_gpo_access_control' = 'enforcing' and
           'ad_site' = "New-Site"
        5. Start SSSD
    :steps:
        1. Authenticate as 'user1' with ssh
        2. Authenticate as 'deny_user1' with ssh
    :expectedresults:
        1. 'user1' authentication is successful
        2. 'deny_user1' authentication is unsuccessful
    :customerscenario: True
    """
    user1 = provider.user("user1").add()
    deny_user1 = provider.user("deny_user1").add()
    provider.site("New-Site").add()

    provider.gpo("site policy").add().policy(
        {
            "SeInteractiveLogonRight": [user1, provider.group("Domain Admins")],
            "SeDenyInteractiveLogonRight": [deny_user1],
        }
    ).link(target=f"cn=New-Site,cn=sites,cn=configuration,{provider.naming_context}")

    client.sssd.domain["ad_gpo_access_control"] = "enforcing"
    client.sssd.domain["ad_site"] = "New-Site"
    client.sssd.start()

    assert client.auth.parametrize(method).password(
        "user1", password="Secret123"
    ), "Allowed user authentication failed!"

    assert not client.auth.parametrize(method).password(
        "deny_user1", password="Secret123"
    ), "Denied user authenticated successfully!"


@pytest.mark.importance("critical")
@pytest.mark.parametrize("method", ["ssh", "su"])
@pytest.mark.topology(KnownTopology.AD)
@pytest.mark.ticket(gh=905)
def test_gpo__only_needs_host_security_filters_and_permissions(client: Client, ad: AD, method: str):
    """
    :title: Group policy object host based access control only needs host security filters and permissions.
    :description:
        GPOs have additional security filters for which objects can apply the policy. The default value is
        'Authenticated Users' group, when tightening security this may be changed. Commonly the computer
        object can only apply the policy.
    :setup:
        1. Create the following users; 'user1', 'deny_user1'
        2. Create the GPO 'site policy' and add 'user1', group and 'Domain Admins' to
           SeInteractiveLogonRight key. Add 'deny_user1' to SeDenyInteractiveLogonRight key
        3. Configure sssd.conf with 'ad_gpo_access_control' = 'enforcing'
        4. Add the permissions so the client has 'gpoapply' permissions,
           and set 'authenticated users' to 'gporead'
        5. Start SSSD
    :steps:
        1. Authenticate as 'user1'
        2. Authenticate as 'deny_user1'
    :expectedresults:
        1. 'user1' authentication is successful
        2. 'deny_user1' authentication is unsuccessful
    :customerscenario: True
    """
    user1 = ad.user("user1").add()
    deny_user1 = ad.user("deny_user1").add()
    computer = client.host.hostname.split(".")[0].upper()

    computer_policy = (
        ad.gpo("computer policy")
        .add()
        .policy(
            {
                "SeInteractiveLogonRight": [user1, ad.group("Domain Admins")],
                "SeDenyInteractiveLogonRight": [deny_user1],
            }
        )
        .link()
    )

    computer_policy.permissions(f"{computer}$", "GpoApply", "Computer")
    computer_policy.permissions("Authenticated Users", "GpoRead", "Group")

    client.sssd.domain["ad_gpo_access_control"] = "enforcing"
    client.sssd.start()

    assert client.auth.parametrize(method).password(
        "user1", password="Secret123"
    ), "Allowed user authentication failed!"

    assert not client.auth.parametrize(method).password(
        "deny_user1", password="Secret123"
    ), "Denied user authenticated successfully!"


@pytest.mark.importance("critical")
@pytest.mark.parametrize("method", ["ssh", "su"])
@pytest.mark.topology(KnownTopologyGroup.AnyAD)
@pytest.mark.ticket(bz=1316164)
def test_gpo__ignores_invalid_and_unnecessary_keys_and_values(
    client: Client, provider: GenericADProvider, method: str
):
    """
    :title: Group policy object host based access control ignores invalid and unnecessary keys and values.
    :description:
        The GPO security database can contain additional keys and keys with empty values for other
        applications. SSSD should only process the relevant keys and ignore the rest. When creating
        the policy bogus keys are added.
    :setup:
        1. Create the following users; 'user1', 'deny_user1'
        2. Create the GPO 'site policy' and add 'user1', group and 'Domain Admins' to
           SeInteractiveLogonRight key. Add 'deny_user1' to SeDenyInteractiveLogonRight key,
           additionally add some empty keys and invalid values to the policy
        3. Configure sssd.conf with 'ad_gpo_access_control' = 'enforcing'
        4. Start SSSD
    :steps:
        1. Authenticate as 'user1'
        2. Authenticate as 'deny_user1'
    :expectedresults:
        1. 'user1' authentication is successful
        2. 'deny_user1' authentication is unsuccessful
    :customerscenario: True
    """
    user1 = provider.user("user1").add()
    deny_user1 = provider.user("deny_user1").add()

    provider.gpo("policy invalid keys and values").add().policy(
        {
            "SeInteractiveLogonRight": [user1, provider.group("Domain Admins")],
            "SeDenyInteractiveLogonRight": [deny_user1],
        },
        cfg={"Service General Setting": {"BITS": "2", "wuaserv": "2", "MpsSvc": "2"}},
    ).link()

    client.sssd.domain["ad_gpo_access_control"] = "enforcing"
    client.sssd.start()

    assert client.auth.parametrize(method).password(
        "user1", password="Secret123"
    ), "Allowed user authentication failed!"

    assert not client.auth.parametrize(method).password(
        "deny_user1", password="Secret123"
    ), "Denied user authenticated successfully!"


@pytest.mark.importance("critical")
@pytest.mark.parametrize("method", ["ssh", "su"])
@pytest.mark.topology(KnownTopology.AD)
@pytest.mark.ticket(bz=1855281)
def test_gpo__skips_unreadable_gpo_policies(client: Client, ad: AD, method: str):
    """
    :title: Group policy object host based access control skips unreadable GPO policies.
    :description:
        When a policy cannot be read, the default behavior is to stop processing all policies,
        to air on the side of caution. Enabling 'ad_gpo_ignore_unreadable' will skip the unreadable
        policies and process the policies that are readable.
    :setup:
        1. Create the following users; 'user1', 'deny_user1'
        2. Create and link the GPO 'site policy' and add 'user1', group and 'Domain Admins' to
           SeInteractiveLogonRight key. Add 'deny_user1 to SeDenyInteractiveLogonRight key
        3. Create and link the GPO 'unreadable policy', and set the permissions for
           'Authenticated Users' to 'None'
        4. Configure sssd.conf with 'ad_gpo_access_control' = 'enforcing' and
           'ad_gpo_ignore_unreadable' = 'True
        5. Start SSSD
    :steps:
        1. Authenticate as 'user1'
        2. Authenticate as 'deny_user1'
    :expectedresults:
        1. 'user1' authentication is successful
        2. 'deny_user1' authentication is unsuccessful
    :customerscenario: True
    """
    user1 = ad.user("user1").add()
    deny_user1 = ad.user("deny_user1").add()

    ad.gpo("site policy").add().policy(
        {
            "SeInteractiveLogonRight": [user1, ad.group("Domain Admins")],
            "SeDenyInteractiveLogonRight": [deny_user1],
        }
    ).link()

    ad.gpo("unreadable policy").add().link().permissions("Authenticated Users", "None", "Group")

    client.sssd.domain["ad_gpo_access_control"] = "enforcing"
    client.sssd.domain["ad_gpo_ignore_unreadable"] = "True"
    client.sssd.start()

    assert client.auth.parametrize(method).password(
        "user1", password="Secret123"
    ), "Allowed user authentication failed!"

    assert not client.auth.parametrize(method).password(
        "deny_user1", password="Secret123"
    ), "Denied user authenticated successfully!"


@pytest.mark.importance("critical")
@pytest.mark.parametrize("method", ["ssh", "su"])
@pytest.mark.topology(KnownTopologyGroup.AnyAD)
@pytest.mark.ticket(bz=2151450)
def test_gpo__finds_all_groups_when_auto_private_groups_is_set_true(
    client: Client, provider: GenericADProvider, method: str
):
    """
    :title: Primary group is missing from users when auto_private_groups are enabled
    :description:
        This tests for a bug where the primary group is not returned when the user is looked up.
    :setup:
        1. Create the following user 'user1'
        2. Create and link the GPO 'site policy' and add 'user1' and 'Domain Admins' to SeInteractiveLogonRight key.
        3. Configure sssd.conf with 'ad_gpo_access_control = enforcing', 'ldap_use_tokengroups = false' and
           'auto_private_groups = true'
        4. Start SSSD
    :steps:
        1. Authenticate as 'user1'
        2. Lookup user
    :expectedresults:
        1. Authentication is successful
        2. User found and primary group 'Domain Users' is listed
    :customerscenario: True
    """
    user1 = provider.user("user1").add()

    provider.gpo("site policy").add().policy(
        {
            "SeInteractiveLogonRight": [user1, provider.group("Domain Admins")],
            "SeDenyInteractiveLogonRight": [],
        }
    ).link()

    client.sssd.domain["ad_gpo_access_control"] = "enforcing"
    client.sssd.domain["auto_private_groups"] = "true"
    client.sssd.domain["ldap_use_tokengroups"] = "false"
    client.sssd.start()

    assert client.auth.parametrize(method).password(
        "user1", password="Secret123"
    ), "Allowed user authentication failed!"

    result = client.tools.id("user1")
    assert result is not None, "User not found!"
    assert result.memberof("domain users"), "User missing from group 'domain users'!"


@pytest.mark.importance("critical")
@pytest.mark.parametrize("method", ["ssh", "su"])
@pytest.mark.parametrize("auto_private_groups", ["true", "false", "hybrid"])
@pytest.mark.topology(KnownTopologyGroup.AnyAD)
@pytest.mark.ticket(gh=7452)
def test_gpo__works_when_auto_private_group_is_used_with_posix_accounts(
    client: Client, provider: GenericADProvider, method: str, auto_private_groups: str
):
    """
    :title: GPO evaluation fails when auto_private_groups used with posix accounts
    :setup:
        1. Create the following user 'user1' and 'deny_user1' with uids and gids
        2. Create and link the GPO 'site policy' and add 'user1' and 'Domain Admins' to
           SeInteractiveLogonRight key. Add 'deny_user1 to SeDenyInteractiveLogonRight key'
        3. Configure sssd.conf with 'ad_gpo_access_control = enforcing',
           'auto_private_groups = parameter' and 'ldap_id_mapping = false'
        4. Start SSSD
    :steps:
        1. Authenticate as 'user1'
        2. Authenticate as 'deny_user1'
    :expectedresults:
        1. Authentication is successful
        2. Authenticated user is unsuccessful
    :customerscenario: True
    """
    user1 = provider.user("user1").add(uid=10000, gid=10000)
    deny_user1 = provider.user("deny_user1").add(uid=10001, gid=10001)

    provider.gpo("site policy").add().policy(
        {
            "SeInteractiveLogonRight": [user1, provider.group("Domain Admins")],
            "SeDenyInteractiveLogonRight": [deny_user1],
        }
    ).link()

    client.sssd.domain["ad_gpo_access_control"] = "enforcing"
    client.sssd.domain["auto_private_groups"] = auto_private_groups
    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    assert client.auth.parametrize(method).password(
        "user1", password="Secret123"
    ), "Allowed user authentication failed!"
    assert not client.auth.parametrize(method).password("deny_user1", password="Secret123"), "Denied user logged in!"


@pytest.mark.importance("critical")
@pytest.mark.parametrize("method", ["ssh", "su"])
@pytest.mark.topology(KnownTopologyGroup.AnyAD)
@pytest.mark.ticket(gh=7591)
def test_gpo__ldap_user_name_attribute_mapping(client: Client, provider: GenericADProvider, method: str):
    """
    :title: GPO evaluation fails when the LDAP attribute "name" is used instead of default sAMAccountName attribute
    :description: The name attribute is not populated in Samba, limiting the test to AD
    :setup:
        1. Create the following user 'user1' and 'deny_user1' with uids and gids
        2. Create and link the GPO 'site policy' and add 'user1' and 'Domain Admins' to
           SeInteractiveLogonRight key. Add 'deny_user1 to SeDenyInteractiveLogonRight key'
        3. Configure sssd.conf with 'ad_gpo_access_control = enforcing',
           'auto_private_groups = false', 'ldap_user_name = name' and 'ldap_id_mapping = false'
        4. Start SSSD
    :steps:
        1. Authenticate as 'user1'
        2. Authenticate as 'deny_user1'
    :expectedresults:
        1. Authentication is successful
        2. Authenticated user is unsuccessful
    :customerscenario: True
    """
    user1 = provider.user("user1").add(uid=10000, gid=10000)
    deny_user1 = provider.user("deny_user1").add(uid=10001, gid=10001)

    provider.gpo("site policy").add().policy(
        {
            "SeInteractiveLogonRight": [user1, provider.group("Domain Admins")],
            "SeDenyInteractiveLogonRight": [deny_user1],
        }
    ).link()

    if isinstance(provider, Samba):
        client.sssd.domain["ldap_user_name"] = "givenName"
    if isinstance(provider, AD):
        client.sssd.domain["ldap_user_name"] = "name"

    client.sssd.domain["ad_gpo_access_control"] = "enforcing"
    client.sssd.domain["auto_private_groups"] = "false"
    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    assert client.auth.parametrize(method).password(
        "user1", password="Secret123"
    ), "Allowed user authentication failed!"
    assert not client.auth.parametrize(method).password("deny_user1", password="Secret123"), "Denied user logged in!"



# Constants
ds_instance_name = 'example1'
ds_instance_name1 = 'example2'
ds_instance_name2 = 'example3'
ds_suffix = 'dc=example,dc=test'
krb_realm = 'EXAMPLE.TEST'
ds_rootdn = 'cn=Directory Manager'
ds_rootpw = 'Secret123'



"""
Identity Tests

These tests cover all the searches, queries, and lookups performed by SSSD.

:requirement: Identity
"""

from __future__ import annotations

import pytest
from sssd_test_framework.roles.client import Client
from sssd_test_framework.roles.generic import GenericADProvider, GenericProvider
from sssd_test_framework.roles.ipa import IPA
from sssd_test_framework.roles.ldap import LDAP
from sssd_test_framework.topology import KnownTopology, KnownTopologyGroup


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
@pytest.mark.parametrize("sssd_service_user", ("root", "sssd"))
@pytest.mark.require(
    lambda client, sssd_service_user: ((sssd_service_user == "root") or client.features["non-privileged"]),
    "SSSD was built without support for running under non-root",
)
def test_identity__lookup_username_with_id_command(client: Client, provider: GenericProvider, sssd_service_user: str):
    """
    :title: Resolve user by name with "id"
    :setup:
        1. Create the following users 'user1', 'user2' and 'user3' specifying the UIDs
        2. Configure SSSD with "ldap_id_mapping = false" and start SSSD
    :steps:
        1. Lookup 'user1', 'user2' and 'user3' using the UID
        2. Check the results
    :expectedresults:
        1. Users are found
        2. Results have the correct name and UIDs
    :customerscenario: False
    """
    ids = [("user1", 10001), ("user2", 10002), ("user3", 10003)]
    for user, id in ids:
        provider.user(user).add(uid=id, gid=id + 500)

    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start(service_user=sssd_service_user)

    for name, uid in ids:
        result = client.tools.id(name)
        assert result is not None, f"User {name} was not found using id!"
        assert result.user.name == name, f"Username {result.user.name} is incorrect, {name} expected!"
        assert result.user.id == uid, f"User id {result.user.id} is incorrect, {uid} expected!"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
@pytest.mark.parametrize("sssd_service_user", ("root", "sssd"))
@pytest.mark.require(
    lambda client, sssd_service_user: ((sssd_service_user == "root") or client.features["non-privileged"]),
    "SSSD was built without support for running under non-root",
)
def test_identity__lookup_uid_with_id_command(client: Client, provider: GenericProvider, sssd_service_user: str):
    """
    :title: Resolve user by uid with "id"
    :setup:
        1. Create the following users 'user1', 'user2' and 'user3' specifying the UIDs and GIDs
        2. Configure SSSD with "ldap_id_mapping = false" and start SSSD
    :steps:
        1. Lookup 'user1', 'user2' and 'user3' using the UID
        2. Check the results
    :expectedresults:
        1. Users are found
        2. Results have the correct name and UID
    :customerscenario: False
    """
    ids = [("user1", 10001), ("user2", 10002), ("user3", 10003)]
    for user, id in ids:
        provider.user(user).add(uid=id, gid=id + 500)

    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start(service_user=sssd_service_user)

    for name, uid in ids:
        result = client.tools.id(uid)
        assert result is not None, f"User with uid {uid} was not found using id!"
        assert result.user.name == name, f"Username {result.user.name} is incorrect, {name} expected!"
        assert result.user.id == uid, f"User id {result.user.id} is incorrect, {uid} expected!"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_identity__lookup_groupname_with_getent(client: Client, provider: GenericProvider):
    """
    :title: Resolve group by name with getent
    :setup:
        1. Create the following groups 'group1', 'group2' and 'group3' specifying the GIDs
        2. Configure SSSD with "ldap_id_mapping = false" and start SSSD
    :steps:
        1. Lookup the groups
        2. Check the results
    :expectedresults:
        1. Groups are found
        2. Results have the correct names and GIDs
    :customerscenario: False
    """
    ids = [("group1", 10001), ("group2", 10002), ("group3", 10003)]
    for group, id in ids:
        provider.group(group).add(gid=id)

    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    for name, gid in ids:
        result = client.tools.getent.group(name)
        assert result is not None, f"Group {name} was not found using getent!"
        assert result.name == name, f"Groupname {result.name} is incorrect, {name} expected!"
        assert result.gid == gid, f"Group gid {result.gid} is incorrect, {gid} expected!"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_identity__lookup_group_gid_with_getent(client: Client, provider: GenericProvider):
    """
    :title: Resolve group with by gid with getent
    :setup:
        1. Create the following groups 'group1', 'group2' and 'group3' specifying the GIDs
        2. Configure SSSD with "ldap_id_mapping = false" and start SSSD
    :steps:
        1. Lookup the groups using their GID with getent
        2. Check the results
    :expectedresults:
        1. Groups are found
        2. Groups have the correct names and GIDs
    :customerscenario: False
    """
    ids = [("group1", 10001), ("group2", 10002), ("group3", 10003)]
    for group, id in ids:
        provider.group(group).add(gid=id)

    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    for name, gid in ids:
        result = client.tools.getent.group(gid)
        assert result is not None, f"Group with gid {gid} was not found using getent!"
        assert result.name == name, f"Groupname {result.name} is incorrect, {name} expected!"
        assert result.gid == gid, f"Group gid {result.gid} is incorrect, {gid} expected!"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_identity__lookup_user_with_getent(client: Client, provider: GenericProvider):
    """
    :title: Resolve user with getent
    :setup:
        1. Create the following users 'user1', 'user2' and 'user3' specifying the UIDs
        2. Configure SSSD with "ldap_id_mapping = false" and start SSSD
    :steps:
        1. Lookup the users using their name
        2. Check the results
        3. Lookup the users using their uid
        4. Check the results
    :expectedresults:
        1. Users are found
        2. Users have correct names and uids
        3. Users are found
        4. Users have correct names and uids
    :customerscenario: False
    """
    ids = [("user1", 10001), ("user2", 10002), ("user3", 10003)]
    for user, id in ids:
        provider.user(user).add(uid=id, gid=id + 500)

    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    for name, uid in ids:
        result = client.tools.getent.passwd(name)
        assert result is not None, f"User {name} was not found using getent!"
        assert result.name == name, f"Username {result.name} is incorrect, {name} expected!"
        assert result.uid == uid, f"User id {result.uid} is incorrect, {uid} expected!"

        result = client.tools.getent.passwd(uid)
        assert result is not None, f"User with uid {uid} was not found using getent!"
        assert result.name == name, f"Username {result.name} is incorrect, {name} expected!"
        assert result.uid == uid, f"User id {result.uid} is incorrect, {uid} expected!"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_identity__lookup_groups_by_name_and_gid_with_getent(client: Client, provider: GenericProvider):
    """
    :title: Resolve groups with getent
    :setup:
        1. Create the following groups 'group1', 'group2' and 'group3' specifying the GIDs
        2. Configure SSSD with "ldap_id_mapping = false" and start SSSD
    :steps:
        1. Lookup the groups using their name
        2. Check the results
        3. Lookup the groups using their gid
        4. Check the results
    :expectedresults:
        1. Groups are found
        2. Groups have correct names and gids
        3. Groups are found
        4. Groups have correct names and gids
    :customerscenario: False
    """
    groups = [("group1", 10001), ("group2", 10002), ("group3", 10003)]
    for group, id in groups:
        provider.group(group).add(gid=id)

    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    for group, id in groups:
        result = client.tools.getent.group(group)
        assert result is not None, f"Group {group} was not found using getent!"
        assert result.name == group, f"Groupname {result.name} is incorrect, {group} expected!"
        assert result.gid == id, f"Group gid {result.gid} is incorrect, {id} expected!"

        result = client.tools.getent.group(id)
        assert result is not None, f"Group with gid {id} was not found using getent!"
        assert result.name == group, f"Groupname {result.name} is incorrect, {group} expected!"
        assert result.gid == id, f"Group gid {result.gid} is incorrect, {id} expected!"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
@pytest.mark.parametrize("sssd_service_user", ("root", "sssd"))
@pytest.mark.require(
    lambda client, sssd_service_user: ((sssd_service_user == "root") or client.features["non-privileged"]),
    "SSSD was built without support for running under non-root",
)
def test_identity__lookup_group_membership_by_username_with_id_command(
    client: Client, provider: GenericProvider, sssd_service_user: str
):
    """
    :title: Check membership of user by group name with "id"
    :setup:
        1. Create the following users 'user1', 'user2', 'user3', and 'user4'
        2. Create 'group1' and add all users to group except for 'user4'
        3. Configure SSSD with "ldap_id_mapping = false" and start SSSD
    :steps:
        1. Lookup users by name
        2. Check results
    :expectedresults:
        1. Users are found
        2. All users except 'user4' are members of 'group1'
    :customerscenario: False
    """
    users = [("user1", "group1"), ("user2", "group1"), ("user3", "group1")]
    u1 = provider.user("user1").add()
    u2 = provider.user("user2").add()
    u3 = provider.user("user3").add()
    u4 = provider.user("user4").add()

    provider.group("group1").add().add_members([u1, u2, u3])

    client.sssd.start(service_user=sssd_service_user)

    for name, groups in users:
        result = client.tools.id(name)
        assert result is not None, f"User {name} was not found using id!"
        assert result.memberof(groups), f"User {name} is a member of the wrong groups!"

    result = client.tools.id(u4.name)
    assert result is not None, "User not found!"
    assert not result.memberof("group1"), f"User {u4.name} is a member of the wrong groups!"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_identity__lookup_group_membership_by_group_with_id_command(client: Client, provider: GenericProvider):
    """
    :title: Check membership of user by gid with "id"
    :setup:
        1. Create the following users 'user1', 'user2' and 'user3' specifying the UIDs and GIDs
        2. Create 'group1' and add all users to group
        3. Configure SSSD with "ldap_id_mapping = false" and start SSSD
    :steps:
        1. Lookup users by name with "id"
        2. Check results
    :expectedresults:
        1. Users are found
        2. Users are members of the group checked using the GID
    :customerscenario: False
    """
    users = [("user1", 1001), ("user2", 1001), ("user3", 1001)]
    u1 = provider.user("user1").add(uid=10001, gid=19001)
    u2 = provider.user("user2").add(uid=10002, gid=19002)
    u3 = provider.user("user3").add(uid=10003, gid=19003)

    provider.group("group1").add(gid=1001).add_members([u1, u2, u3])

    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    for name, gids in users:
        result = client.tools.id(name)
        assert result is not None, f"User {name} was not found using id!"
        assert result.memberof(gids), f"User {name} is member of wrong groups!"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_identity__lookup_initgroups_with_getent(client: Client, provider: GenericProvider):
    """
    :title: Check initgroups of user
    :setup:
        1. Create the following users 'user1', 'user2' and 'user3' specifying the UIDs and GIDs
        2. Create the following groups 'group1', 'group2' and 'group3' and add all users to all groups
        3. Configure SSSD with "ldap_id_mapping = false" and start SSSD
    :steps:
        1. Lookup users using initgroups with getent
        2. Check results
    :expectedresults:
        1. Users are found
        2. Users are in the groups checked by GIDs
    :customerscenario: False
    """
    users = ["user1", "user2", "user3"]
    u1 = provider.user("user1").add(uid=10001, gid=19001)
    u2 = provider.user("user2").add(uid=10002, gid=19002)
    u3 = provider.user("user3").add(uid=10003, gid=19003)

    provider.group("group1").add(gid=10001).add_members([u1, u2, u3])
    provider.group("group2").add(gid=10002).add_members([u1, u2, u3])
    provider.group("group3").add(gid=10003).add_members([u1, u2, u3])

    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    for name in users:
        result = client.tools.getent.initgroups(name)
        assert result.name == name, f"Username {result.name} is incorrect, {name} expected!"
        assert result.memberof([10001, 10002, 10003]), f"User {name} is member of wrong groups!"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_identity__lookup_users_with_fully_qualified_name(client: Client, provider: GenericProvider):
    """
    :title: Resolve user when 'use_fully_qualified_names' is 'true'
    :setup:
        1. Create the following users 'user1', 'user2' and 'user3' specifying the UIDs and GIDs
        2. Configure SSSD with "ldap_id_mapping = false" and "use_fully_qualified_name = true" and start SSSD
    :steps:
        1. Lookup users with their username
        2. Lookup users with their fully qualified name
        3. Check results
    :expectedresults:
        1. Users are not found
        2. Users are found
        3. Users have the correct names and UIDs
    :customerscenario: False
    """
    provider.user("user1").add(uid=10001, gid=19001)
    provider.user("user2").add(uid=10002, gid=19002)

    client.sssd.domain["use_fully_qualified_names"] = "true"
    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    assert client.tools.id("user1") is None, "User user1 should be found only with fq name!"
    assert client.tools.id("user2") is None, "User user2 should be found only with fq name!"

    result = client.tools.id("user1@test")
    assert result is not None, "User user1@test was not found using id!"
    assert result.user.name == "user1@test", f"Username {result.user.name} is incorrect, user1@test expected!"
    assert result.user.id == 10001, f"User id {result.user.id} is incorrect, 10001 expected!"

    result = client.tools.id("user2@test")
    assert result is not None, "User user2@test was not found using id!"
    assert result.user.name == "user2@test", f"Username {result.user.name} is incorrect, user2@test expected!"
    assert result.user.id == 10002, f"User id {result.user.id} is incorrect, 10002 expected!"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_identity__lookup_users_when_case_sensitive_is_false(client: Client, provider: GenericProvider):
    """
    :title: Search user with case-sensitivity is false
    :setup:
        1. Create the following users 'user1', 'user2' and 'user3' specifying the UIDs and GIDs
        2. Configure SSSD with "ldap_id_mapping = false" and "case_sensitive = false" and start SSSD
    :steps:
        1. Lookup users by their name randomizing the capitalization of letters in the name
        2. Check the results
    :expectedresults:
        1. Users are found
        2. Results have the correct name and UID
    :customerscenario: False
    """
    provider.user("user1").add(uid=10001, gid=19001)
    provider.user("user2").add(uid=10002, gid=19002)
    provider.user("user3").add(uid=10003, gid=19003)

    client.sssd.domain["case_sensitive"] = "false"
    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    for name, uid in [
        ("uSer1", 10001),
        ("user1", 10001),
        ("uSER1", 10001),
        ("USEr2", 10002),
        ("uSEr2", 10002),
        ("usER2", 10002),
        ("USer3", 10003),
        ("uSer3", 10003),
        ("USER3", 10003),
    ]:
        result = client.tools.id(name)
        assert result is not None, f"User {name} was not found using id!"
        assert result.user.name == name.lower(), f"Username {result.user.name} is incorrect, {name.lower()} expected!"
        assert result.user.id == uid, f"User id {result.user.id} is incorrect, {uid} expected!"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_identity__lookup_users_fully_qualified_name_and_case_sensitive_is_false(
    client: Client, provider: GenericProvider
):
    """
    :title: Search user with fully qualified name when case-sensitive is false
    :setup:
        1. Create the following users 'user1', 'user2' and 'user3' specifying the UIDs and GIDs
        2. Create the following groups 'group1', 'group2' and 'group3' specifying the GIDs
        3. Configure SSSD with "ldap_id_mapping = false", "case_sensitive = false" and
           "use_fully_qualified_name = true" and start SSSD
    :steps:
        1. Lookup users by their name with id
        2. Lookup users with their fully qualified name randomizing the capitalization of letters in the name
        3. Check that users have correct groups
    :expectedresults:
        1. Users are not found
        2. Users are found
        3. Users are in the correct groups
    :customerscenario: False
    """
    u1 = provider.user("user1").add(gid=101, uid=10001)
    u2 = provider.user("user2").add(gid=102, uid=10002)
    u3 = provider.user("user3").add(gid=103, uid=10003)

    provider.group("group1").add(gid=1001).add_members([u1])
    provider.group("group2").add(gid=1002).add_members([u1, u2])
    provider.group("group3").add(gid=1003).add_members([u1, u2, u3])

    client.sssd.domain["use_fully_qualified_names"] = "true"
    client.sssd.domain["case_sensitive"] = "false"
    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    assert client.tools.id("user1") is None, "User user1 should be found only with fully qualified name!"
    assert client.tools.id("user2") is None, "User user2 should be found only with fully qualified name!"
    assert client.tools.id("user3") is None, "User user3 should be found only with fully qualified name!"

    for name in ["User1@TesT", "UseR1@TesT", "UsER1@TesT"]:
        result = client.tools.id(name)
        assert result is not None, f"User {name} was not found using id!"
        assert result.memberof([101, 1001, 1002, 1003]), f"User {name} is a member of the wrong groups!"

    for name in ["uSer2@TeST", "user2@TEsT", "uSER2@tesT"]:
        result = client.tools.id(name)
        assert result is not None, f"User {name} was not found using id!"
        assert result.memberof([102, 1002, 1003]), f"User {name} is a member of the wrong groups!"
        assert not result.memberof(1001), f"User {name} is in the wrong groups!"

    for name in ["USer3@TeST", "uSer3@TeST", "USER3@Test"]:
        result = client.tools.id(name)
        assert result is not None, f"User {name} was not found using id!"
        assert result.memberof([103, 1003]), f"User {name} is a member of the wrong groups!"
        assert not result.memberof([1001, 1002]), f"User {name} is in the wrong groups!"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopologyGroup.AnyAD)
def test_identity__lookup_id_mapping_of_posix_and_non_posix_user_and_group(
    client: Client, provider: GenericADProvider
):
    """
    :title: Check ID mapping of POSIX and non-POSIX users in AD directories when id mapping is false
    :note: This is a generic provider test, AD is a workaround to create users with no posix attributes
    :setup:
        1. Create user with POSIX attributes
        2. Create group with POSIX attributes
        3. Create user with no POSIX attributes
        4. Create group with no POSIX attributes
        5. Configure SSSD with "ldap_id_mapping = false" and start SSSD
    :steps:
        1. Query POSIX user information
        2. Query Non-POSIX user information
        3. Query POSIX group information
        4. Query Non-POSIX group information
    :expectedresults:
        1. POSIX user found with the correct values
        2. User is not found
        3. POSIX group found with the correct values
        4. Group is not found
    :customerscenario: False
    """

    u1 = provider.user("posix_user").add(
        uid=10001, gid=20001, password="Secret123", gecos="User for tests", shell="/bin/bash"
    )
    provider.group("posix_group").add(gid=20001).add_member(u1)

    u2 = provider.user("nonposix_user").add(password="Secret123")
    provider.group("nonposix_group").add().add_member(u2)

    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    user = client.tools.getent.passwd("posix_user")
    assert user is not None, "posix-user not found!"
    assert user.uid == 10001, "gid returned does not matched the one provided!"
    assert user.gid == 20001, "uid returned does not matched the one provided!"
    assert client.tools.getent.passwd("nonposix_user") is None, "nonposix-user found!"

    group = client.tools.getent.group("posix_group")
    assert group is not None, "posix-group not found!"
    assert group.gid == 20001, "gid is not the correct value!"
    assert client.tools.getent.group("nonposix_group") is None, "nonposix-group found!"


@pytest.mark.ticket(bz=1695577)
@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_identity__lookup_when_auto_private_groups_is_set_to_true(client: Client, provider: GenericProvider):
    """
    :title: Look up users when auto private groups is set to true
    :description:
        When true, the user's gid will match the uid, even when the object isn't
        a real group in the directory. If it is a real group, the user values will override
        the name and gid. In this test, only the users exist in LDAP.
    :setup:
        1. Create the following users 'user_same_gid', 'user_diff_gid' and 'user_group_gid'.
            Set the uids and gids to match the condition defined by in the username
        2. Configure SSSD with "ldap_id_mapping = false" and "auto_private_groups = true" and start SSSD
    :steps:
        1. Lookup all users and compare their uid to their gid
        2. Lookup up the user's gid
    :expectedresults:
        1. All users uid and gid match
        2. All groups are found
    :customerscenario: True
    :requirement: IDM-SSSD-REQ: SSSD can automatically create user private groups for users
    """
    provider.user("user_same_gid").add(uid=111111, gid=111111)
    provider.user("user_diff_gid").add(uid=222222, gid=333333)
    provider.user("user_no_gid").add(uid=444444)

    client.sssd.domain["auto_private_groups"] = "true"
    client.sssd.domain["ldap_id_mapping"] = "false"

    client.sssd.start()

    for i in [("user_same_gid", 111111), ("user_diff_gid", 222222), ("user_no_gid", 444444)]:
        user = client.tools.getent.passwd(i[0])
        assert user is not None, f"User '{i[0]}' is not found!"
        assert user.uid == (i[1]), "uid does not match expected value!"
        assert user.uid == user.gid, "uid does not match gid!"

        group = client.tools.getent.group(i[0])
        assert group is not None, f"{i[0]} group is not found!"


@pytest.mark.ticket(bz=1695577)
@pytest.mark.importance("high")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_identity__lookup_when_auto_private_groups_is_set_to_false(client: Client, provider: GenericProvider):
    """
    :title: Look up users when auto private groups is set to false
    :description:
        When false, to be able to look up the group, the group needs to exist
        in the directory. In this test, only the 'user_group_gid' user has a valid LDAP group.
    :setup:
        1. Create a group
        2. Create the following users 'user_same_gid', 'user_diff_gid' and 'user_group_gid'.
            Set the uids and gids to match the condition defined by in the username
        3. Configure SSSD with "ldap_id_mapping = false" and "auto_private_groups = false" and start SSSD
    :steps:
        1. Lookup 'user_same_gid' and look up the user's gid
        2. Lookup 'user_diff_gid' and look up the user's gid
        3. Lookup 'user_group_gid' and look up the user's gid
    :expectedresults:
        1. The user is found with the configured values, and the group is *NOT* found
        2. The user is found with the configured values, and the group is *NOT* found
        3. The user is found with the configured values, and the group is found
    :customerscenario: True
    :requirement: IDM-SSSD-REQ: SSSD can automatically create user private groups for users
    """
    provider.group("group").add(gid=444444)
    provider.user("user_same_gid").add(uid=111111, gid=111111)
    provider.user("user_diff_gid").add(uid=222222, gid=333333)
    provider.user("user_group_gid").add(uid=444444, gid=444444)

    client.sssd.domain["auto_private_groups"] = "false"
    client.sssd.domain["ldap_id_mapping"] = "false"

    client.sssd.start()

    result = client.tools.getent.passwd("user_same_gid")
    assert result is not None, "User 'user_same_gid' not found!"
    assert result.gid == 111111, "gid does not match expected value!"

    # IPA manages auto_private_groups on the server and is true by default
    if isinstance(provider, IPA):
        assert client.tools.getent.group(111111), "Group is not found!"
    else:
        assert not client.tools.getent.group(111111), "Group should not be found!"

    result = client.tools.getent.passwd("user_diff_gid")
    assert result is not None, "User 'user_diff_gid' is not found!"
    assert result.gid == 333333, "gid does not match expected value!"
    assert client.tools.getent.group(333333) is None, "group is found!"

    result = client.tools.getent.passwd("user_group_gid")
    assert result is not None, "User 'user_group_gid' is not found!"
    assert result.gid == 444444, "gid does not match expected value!"
    assert client.tools.getent.group(444444) is not None, "group is not found!"


@pytest.mark.ticket(bz=1695577)
@pytest.mark.importance("high")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_identity__lookup_when_auto_private_groups_is_set_to_hybrid(client: Client, provider: GenericProvider):
    """
    :title: Look up users when auto private groups is set to hybrid
    :description:
        When set to hybrid, if the uid and gid match for the user, it will act as if a
        group exists similar to when the parameter is set to true. Like when it's set to false, if the
        group exists in LDAP, it will look at whatever is existing for its values.
    :setup:
        1. Create a group
        2. Create the following users 'user_same_gid', 'user_diff_gid', 'user_no_gid' and 'user_group_gid'.
            Set the uids and gids to match the condition defined by in the username
        3. Configure SSSD with "ldap_id_mapping = false" and "auto_private_groups = hybrid" start SSSD
    :steps:
        1. Lookup 'user_same_gid' and look up the user's gid
        2. Lookup 'user_diff_gid' and look up the user's gid
        3. Lookup 'user_no_gid' and look up the user's gid
        4. Lookup 'user_group_gid' and look up the user's gid
    :expectedresults:
        1. The user is found with the configured values, and the group is found
        2. The user is found with the configured values, and the group is *NOT* found
        3. The user is found with the configured values, and the group is *NOT* found
        4. The user is found with the configured values, and the group is found
    :customerscenario: True
    :requirement: IDM-SSSD-REQ: SSSD can automatically create user private groups for users
    """
    provider.group("group").add(gid=55555)
    provider.user("user_same_gid").add(uid=111111, gid=111111)
    provider.user("user_diff_gid").add(uid=222222, gid=333333)
    provider.user("user_no_gid").add(uid=444444)
    provider.user("user_group_gid").add(uid=555555, gid=555555)

    client.sssd.domain["auto_private_groups"] = "hybrid"
    client.sssd.domain["ldap_id_mapping"] = "false"

    client.sssd.start()

    result = client.tools.getent.passwd("user_same_gid")
    assert result is not None, "User 'user_same_gid' not found!"
    assert result.gid == 111111, "gid does not match expected value!"
    assert client.tools.getent.group(111111) is not None, "auto private group not found!"

    result = client.tools.getent.passwd("user_diff_gid")
    assert result is not None, "User 'user_diff_gid' not found!"
    assert result.gid == 333333, "gid does not match expected value!"
    assert client.tools.getent.group(333333) is None, "auto private group should not be found!"

    # IPA manages auto_private_groups on the server and is true by default
    if isinstance(provider, IPA):
        result = client.tools.getent.passwd("user_group_gid")
        assert result is not None, "User 'user_group_gid' not found!"
        assert result.gid == 555555, "gid is not found!"
    else:
        assert client.tools.getent.passwd("user_no_id") is None, "gid should not be found!"

    result = client.tools.getent.passwd("user_group_gid")
    assert result is not None, "User 'user_group_gid' not found!"
    assert result.gid == 555555, "gid does not match expected value!"
    assert client.tools.getent.group(555555) is not None, "auto private group not found!"


@pytest.mark.importance("low")
@pytest.mark.topology(KnownTopology.LDAP)
def test_identity__lookup_when_backend_restarts(client: Client, ldap: LDAP):
    """
    :title: Look up user when backend is restarted with previous lookup unfinished
    :description:
        If there is an active lookup for a user and the backend is restarted
        before this lookup is finished, the next lookup of the same user after
        the restart must not timeout.
    :setup:
        1. Add a user "tuser"
        2. Start SSSD
    :steps:
        1. Add 10s network traffic delay to the LDAP host
        2. Lookup "tuser" asynchronously
        3. Kill sssd_be with SIGKILL so it is restarted
        4. Remove the network traffic delay
        5. Lookup of "tuser" must yield the user and not timeout
    :expectedresults:
        1. Network traffic is delayed
        2. Lookup hangs, does not finish and waits for a timeout
        3. The backend process is restarted
        4. Network traffic is no longer delayed
        5. User lookup returns the user immediately
    :customerscenario: False
    """
    ldap.user("tuser").add()

    client.sssd.start()

    # Add a delay so the next lookup will hang
    client.tc.add_delay(ldap, "10s")
    client.host.conn.async_run("getent passwd tuser")

    # Kill backend and remove the delay
    client.host.conn.run("kill -KILL $(pidof sssd_be)")
    client.tc.remove_delay(ldap)

    # The next lookup should not timeout
    result = client.tools.wait_for_condition("getent passwd tuser", timeout=5)
    assert "tuser" in result.stdout, "tuser was not found"



"""
SSSD Schema Tests.

Tests related to directory schemas, formal definitions of LDAP objectClasses and attributes.

These tests are generic topology and will run against AD, Samba, IPA and LDAP.
Specific topologies test may reside in their corresponding test file.

:requirement: ldap_extra_attrs
"""

from __future__ import annotations

import pytest
from sssd_test_framework.roles.client import Client
from sssd_test_framework.roles.generic import GenericProvider
from sssd_test_framework.topology import KnownTopologyGroup


@pytest.mark.importance("high")
@pytest.mark.ticket(gh=4153, bz=1362023)
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
@pytest.mark.parametrize("attrs", ["mail, firstname:givenname, lastname:sn", "given_email:mail"])
def test_schema__user_extra_attributes_are_populated(client: Client, provider: GenericProvider, attrs: str):
    """
    :title: SSSD starts correctly when ldap_extra_attrs is configured
    :setup:
        1. Create user "user1"
        2. Configure SSSD with "ldap_user_extra_attrs = attribute:value"
    :steps:
        1. Start SSSD
        2. Lookup user
    :expectedresults:
        1. SSSD starts with no errors
        2. User found and name matches
    :customerscenario: False
    """
    provider.user("user1").add()
    client.sssd.domain["ldap_user_extra_attrs"] = attrs

    try:
        client.sssd.start()
    except Exception as e:
        pytest.fail(f"Exception shouldn't be raised but we got {type(e)}: str(e)")

    result = client.tools.getent.passwd("user1")
    assert result is not None, "User not found!"
    assert result.name == "user1", f"User 'user1' name is not the expected value `{result.name}`!"



#!/usr/bin/env python
# coding=utf-8

# Authors:
#   Jakub Hrozek <jhrozek@redhat.com>
#
# Copyright (C) 2009  Red Hat
# see file 'COPYING' for use and warranty information
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import tempfile
import shutil
import unittest
import subprocess
import errno

# module under test
import pysss


class LocalTest(unittest.TestCase):
    local_path = "/var/lib/sss/db/sssd.ldb"

    def setUp(self):
        self.local = pysss.local()

    def _run_and_check(self, runme):
        (status, output) = subprocess.call(runme, shell=True)
        self.failUnlessEqual(status, 0, output)

    def _get_object_info(self, name, subtree, domain):
        search_dn = "dn=name=%s,cn=%s,cn=%s,cn=sysdb" % (name, subtree, domain)
        try:
            cmd = "ldbsearch -H %s %s" % (self.local_path, search_dn)
            output = subprocess.check_call(cmd, shell=True)
            output = output.decode('utf-8')
        except subprocess.CalledProcessError:
            return {}

        kw = {}
        for key, value in \
                [ln.split(':') for ln in output.split('\n') if ":" in ln]:
            kw[key] = value.strip()

        del kw['asq']
        return kw

    def get_user_info(self, name, domain="LOCAL"):
        return self._get_object_info(name, "users", domain)

    def get_group_info(self, name, domain="LOCAL"):
        return self._get_object_info(name, "groups", domain)

    def _validate_object(self, kw, name, **kwargs):
        if kw == {}:
            self.fail("Could not get %s info" % name)
        for key in kwargs.keys():
            self.assert_(str(kwargs[key]) == str(kw[key]),
                         "%s %s != %s %s" % (key, kwargs[key], key, kw[key]))

    def validate_user(self, username, **kwargs):
        return self._validate_object(self.get_user_info(username), "user",
                                     **kwargs)

    def validate_group(self, groupname, **kwargs):
        return self._validate_object(self.get_group_info(groupname), "group",
                                     **kwargs)

    def _validate_no_object(self, kw, name):
        if kw != {}:
            self.fail("Got %s info" % name)

    def validate_no_user(self, username):
        return self._validate_no_object(self.get_user_info(username), "user")

    def validate_no_group(self, groupname):
        return self._validate_no_object(self.get_group_info(groupname),
                                        "group")

    def _get_object_membership(self, name, subtree, domain):
        search_dn = "dn=name=%s,cn=%s,cn=%s,cn=sysdb" % (name, subtree, domain)
        try:
            cmd = "ldbsearch -H %s %s" % (self.local_path, search_dn)
            output = subprocess.check_call(cmd, shell=True)
            output = output.decode('utf-8')
        except subprocess.CalledProcessError:
            return []

        members = [value.strip() for key, value in
                   [ln.split(':') for ln in output.split('\n') if ":" in ln]
                   if key == "memberof"]
        return members

    def _assertMembership(self, name, group_list, subtree, domain):
        members = self._get_object_membership(name, subtree, domain)
        for group in group_list:
            group_dn = "name=%s,cn=groups,cn=%s,cn=sysdb" % (group, domain)
            if group_dn in members:
                members.remove(group_dn)
            else:
                self.fail("Cannot find required group %s" % group_dn)

        if len(members) > 0:
            self.fail("More groups than selected")

    def assertUserMembership(self, name, group_list, domain="LOCAL"):
        return self._assertMembership(name, group_list, "users", domain)

    def assertGroupMembership(self, name, group_list, domain="LOCAL"):
        return self._assertMembership(name, group_list, "groups", domain)

    def get_user_membership(self, name, domain="LOCAL"):
        return self._get_object_membership(name, "users", domain)

    def get_group_membership(self, name, domain="LOCAL"):
        return self._get_object_membership(name, "groups", domain)

    def add_group(self, groupname):
        self._run_and_check("sss_groupadd %s" % (groupname))

    def remove_group(self, groupname):
        self._run_and_check("sss_groupdel %s" % (groupname))

    def add_user(self, username):
        self._run_and_check("sss_useradd %s" % (username))

    def add_user_not_home(self, username):
        self._run_and_check("sss_useradd -M %s" % (username))

    def remove_user(self, username):
        self._run_and_check("sss_userdel %s" % (username))

    def remove_user_not_home(self, username):
        self._run_and_check("sss_userdel -R %s" % (username))


class SanityTest(unittest.TestCase):
    def testInstantiate(self):
        "Test that the local backed binding can be instantiated"
        local = pysss.local()
        self.assert_(local.__class__, "<type 'sss.local'>")


class UseraddTest(LocalTest):
    def tearDown(self):
        if self.username:
            self.remove_user(self.username)

    def testUseradd(self):
        "Test adding a local user"
        self.username = "testUseradd"
        self.local.useradd(self.username)
        self.validate_user(self.username)
        # check home directory was created with default name
        self.assertEquals(os.access("/home/%s" % self.username, os.F_OK), True)

    def testUseraddWithParams(self):
        "Test adding a local user with modified parameters"
        self.username = "testUseraddWithParams"
        self.local.useradd(self.username,
                           gecos="foo bar",
                           homedir="/home/foobar",
                           shell="/bin/zsh")
        self.validate_user(self.username,
                           gecos="foo bar",
                           homeDirectory="/home/foobar",
                           loginShell="/bin/zsh")
        # check home directory was created with nondefault name
        self.assertEquals(os.access("/home/foobar", os.F_OK), True)

    def testUseraddNoHomedir(self):
        "Test adding a local user without creating his home dir"
        self.username = "testUseraddNoHomedir"
        self.local.useradd(self.username, create_home=False)
        self.validate_user(self.username)
        # check home directory was not created
        username_path = "/home/%s" % self.username
        self.assertEquals(os.access(username_path, os.F_OK), False)
        self.local.userdel(self.username, remove=False)
        self.username = None  # fool tearDown into not removing the user

    def testUseraddAlternateSkeldir(self):
        "Test adding a local user and init his homedir from a custom location"
        self.username = "testUseraddAlternateSkeldir"

        skeldir = tempfile.mkdtemp()
        fd, path = tempfile.mkstemp(dir=skeldir)
        fdo = os.fdopen(fd)
        fdo.flush()
        fdo.close
        self.assertEquals(os.access(path, os.F_OK), True)
        filename = os.path.basename(path)

        try:
            self.local.useradd(self.username, skel=skeldir)
            self.validate_user(self.username)
            path = "/home/%s/%s" % (self.username, filename)
            self.assertEquals(os.access(path, os.F_OK), True)
        finally:
            shutil.rmtree(skeldir)

    def testUseraddToGroups(self):
        "Test adding a local user with group membership"
        self.username = "testUseraddToGroups"
        self.add_group("gr1")
        self.add_group("gr2")
        try:
            self.local.useradd(self.username,
                               groups=["gr1", "gr2"])
            self.assertUserMembership(self.username,
                                      ["gr1", "gr2"])
        finally:
            self.remove_group("gr1")
            self.remove_group("gr2")

    def testUseraddWithUID(self):
        "Test adding a local user with a custom UID"
        self.username = "testUseraddWithUID"
        self.local.useradd(self.username,
                           uid=1024)
        self.validate_user(self.username,
                           uidNumber=1024)


class UseraddTestNegative(LocalTest):
    def testUseraddNoParams(self):
        "Test that local.useradd() requires the username parameter"
        self.assertRaises(TypeError, self.local.useradd)

    def testUseraddUserAlreadyExists(self):
        "Test adding a local with a duplicate name"
        self.username = "testUseraddUserAlreadyExists"
        self.local.useradd(self.username)
        try:
            self.local.useradd(self.username)
        except IOError as e:
            self.assertEquals(e.errno, errno.EEXIST)
        else:
            self.fail("Was expecting exception")
        finally:
            self.remove_user(self.username)

    def testUseraddUIDAlreadyExists(self):
        "Test adding a local with a duplicate user ID"
        self.username = "testUseraddUIDAlreadyExists1"
        self.local.useradd(self.username, uid=1025)
        try:
            self.local.useradd("testUseraddUIDAlreadyExists2", uid=1025)
        except IOError as e:
            self.assertEquals(e.errno, errno.EEXIST)
        else:
            self.fail("Was expecting exception")
        finally:
            self.remove_user(self.username)


class UserdelTest(LocalTest):
    def testUserdel(self):
        self.add_user("testUserdel")
        self.assertEquals(os.access("/home/testUserdel", os.F_OK), True)
        self.validate_user("testUserdel")
        self.local.userdel("testUserdel")
        self.validate_no_user("testUserdel")
        self.assertEquals(os.access("/home/testUserdel", os.F_OK), False)

    def testUserdelNotHomedir(self):
        self.add_user("testUserdel")
        self.assertEquals(os.access("/home/testUserdel", os.F_OK), True)
        self.validate_user("testUserdel")
        self.local.userdel("testUserdel", remove=False)
        self.validate_no_user("testUserdel")
        self.assertEquals(os.access("/home/testUserdel", os.F_OK), True)
        shutil.rmtree("/home/testUserdel")
        os.remove("/var/mail/testUserdel")

    def testUserdelNegative(self):
        self.validate_no_user("testUserdelNegative")
        try:
            self.local.userdel("testUserdelNegative")
        except IOError as e:
            self.assertEquals(e.errno, errno.ENOENT)
        else:
            self.fail("Was expecting exception")


class UsermodTest(LocalTest):
    def setUp(self):
        self.local = pysss.local()
        self.username = "UsermodTest"
        self.add_user_not_home(self.username)

    def tearDown(self):
        self.remove_user_not_home(self.username)

    def testUsermod(self):
        "Test modifying user attributes"
        self.local.usermod(self.username,
                           gecos="foo bar",
                           homedir="/home/foobar",
                           shell="/bin/zsh")
        self.validate_user(self.username,
                           gecos="foo bar",
                           homeDirectory="/home/foobar",
                           loginShell="/bin/zsh")

    def testUsermodUID(self):
        "Test modifying UID"
        self.local.usermod(self.username,
                           uid=1024)
        self.validate_user(self.username,
                           uidNumber=1024)

    def testUsermodGroupMembership(self):
        "Test adding to and removing from groups"
        self.add_group("gr1")
        self.add_group("gr2")

        try:
            self.local.usermod(self.username,
                               addgroups=["gr1", "gr2"])
            self.assertUserMembership(self.username,
                                      ["gr1", "gr2"])
            self.local.usermod(self.username,
                               rmgroups=["gr2"])
            self.assertUserMembership(self.username,
                                      ["gr1"])
            self.local.usermod(self.username,
                               rmgroups=["gr1"])
            self.assertUserMembership(self.username,
                                      [])
        finally:
            self.remove_group("gr1")
            self.remove_group("gr2")

    def testUsermodLockUnlock(self):
        "Test locking and unlocking user"
        self.local.usermod(self.username,
                           lock=self.local.lock)
        self.validate_user(self.username,
                           disabled="true")
        self.local.usermod(self.username,
                           lock=self.local.unlock)
        self.validate_user(self.username,
                           disabled="false")


class GroupaddTest(LocalTest):
    def tearDown(self):
        if self.groupname:
            self.remove_group(self.groupname)

    def testGroupadd(self):
        "Test adding a local group"
        self.groupname = "testGroupadd"
        self.local.groupadd(self.groupname)
        self.validate_group(self.groupname)

    def testGroupaddWithGID(self):
        "Test adding a local group with a custom GID"
        self.groupname = "testUseraddWithGID"
        self.local.groupadd(self.groupname,
                            gid=1024)
        self.validate_group(self.groupname,
                            gidNumber=1024)


class GroupaddTestNegative(LocalTest):
    def testGroupaddNoParams(self):
        "Test that local.groupadd() requires the groupname parameter"
        self.assertRaises(TypeError, self.local.groupadd)

    def testGroupaddUserAlreadyExists(self):
        "Test adding a local with a duplicate name"
        self.groupname = "testGroupaddUserAlreadyExists"
        self.local.groupadd(self.groupname)
        try:
            self.local.groupadd(self.groupname)
        except IOError as e:
            self.assertEquals(e.errno, errno.EEXIST)
        else:
            self.fail("Was expecting exception")
        finally:
            self.remove_group(self.groupname)

    def testGroupaddGIDAlreadyExists(self):
        "Test adding a local with a duplicate group ID"
        self.groupname = "testGroupaddGIDAlreadyExists1"
        self.local.groupadd(self.groupname, gid=1025)
        try:
            self.local.groupadd("testGroupaddGIDAlreadyExists2", gid=1025)
        except IOError as e:
            self.assertEquals(e.errno, errno.EEXIST)
        else:
            self.fail("Was expecting exception")
        finally:
            self.remove_group(self.groupname)


class GroupdelTest(LocalTest):
    def testGroupdel(self):
        self.add_group("testGroupdel")
        self.validate_group("testGroupdel")
        self.local.groupdel("testGroupdel")
        self.validate_no_group("testGroupdel")

    def testGroupdelNegative(self):
        self.validate_no_group("testGroupdelNegative")
        try:
            self.local.groupdel("testGroupdelNegative")
        except IOError as e:
            self.assertEquals(e.errno, errno.ENOENT)
        else:
            self.fail("Was expecting exception")


class GroupmodTest(LocalTest):
    def setUp(self):
        self.local = pysss.local()
        self.groupname = "GroupmodTest"
        self.add_group(self.groupname)

    def tearDown(self):
        self.remove_group(self.groupname)

    def testGroupmodGID(self):
        "Test modifying UID"
        self.local.groupmod(self.groupname,
                            gid=1024)
        self.validate_group(self.groupname,
                            gidNumber=1024)

    def testGroupmodGroupMembership(self):
        "Test adding to groups"
        self.add_group("gr1")
        self.add_group("gr2")
        try:
            self.local.groupmod(self.groupname,
                                addgroups=["gr1", "gr2"])
            self.assertGroupMembership(self.groupname,
                                       ["gr1", "gr2"])
            self.local.groupmod(self.groupname,
                                rmgroups=["gr2"])
            self.assertGroupMembership(self.groupname,
                                       ["gr1"])
            self.local.groupmod(self.groupname,
                                rmgroups=["gr1"])
            self.assertGroupMembership(self.groupname,
                                       [])
        finally:
            self.remove_group("gr1")
            self.remove_group("gr2")


# -------------- run the test suite -------------- #
if __name__ == "__main__":
    unittest.main()
